# Check if pandas is available in the current environment
# if it is the case this module can handle Dataframe directly
try:
  import pandas as pd
  has_pandas = True

except:
  has_pandas = False

from epyk.core import html
from epyk.core.css import Defaults


class Lists(object):

  def __init__(self, context):
    self.context = context

  def _recordSet(self, recordSet, column):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param recordSet:
    :param column:
    """
    data = None
    is_converted = False
    if has_pandas:
      if isinstance(recordSet, pd.DataFrame):
        data = [{'name': r, 'value': r} for r in recordSet[column].unique().tolist()]
        is_converted = True

    if not is_converted:
      result = {}
      for rec in recordSet:
        result[rec[column]] = {'name': rec[column], 'value': rec[column]}
      data = [result[k] for k in sorted(result.keys())]
    return data

  def select(self, records=None, htmlCode=None, label=None, selected=None, width=(100, "%"), height=(None, "%"), column=None, filter=None, profile=None, multiple=False, options=None):
    """
    Description:
    ------------
    HTML Select component

    Usage::

      rptObj.ui.select(["A", "B", "C"], label="label", selected="C", multiple=True,
                      options={"title": "ttle", 'showTick': True, 'maxOptions': 2})
      s.selected = "B"
      s.change(rptObj.js.console.log(s.dom.val))

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlSelect.Select`

    Related Pages:

			https://silviomoreto.github.io/bootstrap-select/examples/
    https://www.npmjs.com/package/bootstrap-select-v4
    https://www.jqueryscript.net/form/Bootstrap-4-Dropdown-Select-Plugin-jQuery.html

    Attributes:
    ----------
    :param records: The input data. Can be a list or a dataFrame
    :param htmlCode: Optional. The component identifier code (for bot
    :param label: Optional. The HTML label attached to the component
    :param selected: The selected values
    :param width: Optional. Integer for the component width
    :param height: Optional. Integer for the component height
    :param column:
    :param filter:
    :param profile: Optional. A flag to set the component performance storage
    :param multiple: Boolean. To set if the component can handle multiple selections
    :param options: The select options as defined https://developer.snapappointments.com/bootstrap-select/options/
    """
    records = records or []
    options = {} if options is None else options

    all_selected = options.get("allSelected", False)
    empty_selected = options.get("empty_selected", True)
    if column is not None:
      if filter is not None:
        if filter == True:
          self._filter(records, column, {'allSelected': all_selected, 'operation': options.get("operation", "in")})
      records = self._recordSet(records, column)
    elif isinstance(records, (list, tuple)) and len(records) > 0 and not isinstance(records[0], dict):
      records = [{'name': rec, 'value': rec} for rec in records]
    if all_selected:
      records = [{'name': 'All', 'value': ''}] + records
    if empty_selected:
      records = [{'name': '', 'value': ''}] + records
    if multiple:
      if not isinstance(multiple, dict):
        multiple = {"max": 2}
      if selected is not None:
        for rec in records:
          if rec["value"] in selected:
            rec["selected"] = True
      return html.HtmlSelect.Select(self.context.rptObj, records, htmlCode, width, height, filter, profile, multiple, options)

    if selected is not None:
      for rec in records:
        if rec["value"] == selected:
          rec["selected"] = True
    html_select = html.HtmlSelect.Select(self.context.rptObj, records, htmlCode, width, height, filter, profile, multiple, options)
    return html_select

  def lookup(self, lookupData=None, htmlCode=None, label=None, selected=None, width=(100, "%"), height=(None, "%"), column=None, filter=None, profile=None, multiple=False, options=None):
    """
    Description:
    ------------
    HTML Select component

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlSelect.Lookup`

    Related Pages:

			https://silviomoreto.github.io/bootstrap-select/examples/
    https://www.npmjs.com/package/bootstrap-select-v4
    https://www.jqueryscript.net/form/Bootstrap-4-Dropdown-Select-Plugin-jQuery.html

    Attributes:
    ----------
    :param lookupData: The input data. Can be a list or a dataFrame
    :param htmlCode: Optional. The component identifier code (for bot
    :param label: Optional. The HTML label attached to the component
    :param selected: The selected values
    :param width: Optional. Integer for the component width
    :param height: Optional. Integer for the component height
    :param column:
    :param filter:
    :param profile: Optional. A flag to set the component performance storage
    :param multiple: Boolean. To set if the component can handle multiple selections
    :param options: The select options as defined https://developer.snapappointments.com/bootstrap-select/options/
    """
    options = {} if options is None else options
    html_select = html.HtmlSelect.Lookup(self.context.rptObj, lookupData, htmlCode, width, height, filter, profile, multiple, options)
    return html_select

  def item(self, text=None):
    """
    Description:
    ------------

    Usage::

      l = rptObj.ui.lists.list(["A", "B"])

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlList.List`

    Related Pages:

      https://www.w3schools.com/bootstrap/bootstrap_list_groups.asp
    http://astronautweb.co/snippet/font-awesome/
    """
    html_item = html.HtmlList.Li(self.context.rptObj, text)
    return html_item

  def list(self, data=None, color=None, width=('auto', ""), height=(None, 'px'), htmlCode=None, helper=None,
           options=None, profile=None):
    """
    Description:
    ------------

    Usage::

      l = rptObj.ui.lists.list(["A", "B"])

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlList.List`

    Related Pages:

      https://www.w3schools.com/bootstrap/bootstrap_list_groups.asp
    http://astronautweb.co/snippet/font-awesome/
    """
    html_list = html.HtmlList.List(self.context.rptObj, data or [], color, width, height, htmlCode,
                                   helper, options or {}, profile)
    html_list.css({"list-style": 'none'})
    return html_list

  def items(self, records=None, width=(100, "%"), height=(None, "%"), column=None, options=None, htmlCode=None, profile=None, helper=None):
    """
    """
    if column is not None:
      values = set()
      for rec in records:
        values.add(rec[column])
      records = sorted(list(values))

    html_item = html.HtmlList.Items(self.context.rptObj, 'text', records, width, height, options, htmlCode, profile, helper)
    html_item.css({"list-style-type": 'none'})
    html_item.style.css.padding_left = '15px'
    return html_item

  def numbers(self, data=None, width=('auto', ""), height=(None, 'px'), column=None, htmlCode=None, options=None, profile=None, helper=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.lists.numbers(["A", "B"])

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlList.List`

    Related Pages:

        https://www.w3schools.com/html/html_lists.asp
    https://www.w3.org/wiki/CSS/Properties/list-style-type
    """
    html_list = self.items(data, width, height, column, options, htmlCode, profile, helper)
    html_list.css({"list-style-type": 'decimal'})
    return html_list

  def alpha(self, data=None, width=('auto', ""), height=(None, 'px'), column=None, htmlCode=None, options=None, profile=None, helper=None):
    """
    Description:
    ------------

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlList.List`

    :param data:
    :param color:
    :param width:
    :param height:
    :param htmlCode:
    :param helper:
    :param options:
    :param profile:
    """
    html_list = self.items(data, width, height, column, options, htmlCode, profile, helper)
    html_list.css({"list-style-type": 'lower-alpha'})
    return html_list

  def roman(self, data=None, width=('auto', ""), height=(None, 'px'), column=None, htmlCode=None, options=None, profile=None, helper=None):
    """
    Description:
    ------------

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlList.List`

    :param data:
    :param color:
    :param width:
    :param height:
    :param htmlCode:
    :param helper:
    :param options:
    :param profile:
    :return:
    """
    html_list = self.items(data, width, height, column, options, htmlCode, profile, helper)
    html_list.css({"list-style-type": 'lower-roman'})
    return html_list

  def points(self, data=None, width=('auto', ""), height=(None, 'px'), column=None, htmlCode=None, options=None, profile=None, helper=None):
    """
    Description:
    ------------

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlList.List`

    Related Pages:

			https://www.w3schools.com/html/html_lists.asp

    Attributes:
    ----------
    :param data:
    :param color:
    :param width:
    :param height:
    :param htmlCode:
    :param helper:
    :param options:
    :param profile:
    """
    html_list = self.items(data, width, height, column, options, htmlCode, profile, helper)
    html_list.css({"list-style-type": 'circle'})
    return html_list

  def disc(self, data=None, width=('auto', ""), height=(None, 'px'), column=None, htmlCode=None, helper=None, options=None, profile=None):
    """
    Description:
    ------------

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlList.List`

    Related Pages:

			https://www.w3schools.com/cssref/pr_list-style-type.asp

    Attributes:
    ----------
    :param data:
    :param color:
    :param width:
    :param height:
    :param htmlCode:
    :param helper:
    :param options:
    :param profile:
    """
    html_list = self.items(data, width, height, column, options, htmlCode, profile, helper)
    html_list.css({"list-style-type": 'disc'})
    return html_list

  def squares(self, data=None, width=('auto', ""), height=(None, 'px'), column=None, htmlCode=None, helper=None, options=None, profile=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.lists.squares(["A", "B"])

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlList.List`

    Related Pages:

        https://www.w3schools.com/cssref/pr_list-style-type.asp

    Attributes:
    ----------
    :param data:
    :param color:
    :param width:
    :param height:
    :param htmlCode:
    :param helper:
    :param options:
    :param profile:
    """
    html_list = self.items(data, width, height, column, options, htmlCode, profile, helper)
    html_list.css({"list-style-type": 'square'})
    return html_list

  def groups(self, data=None, categories=None, color=None, width=('auto', ""), height=(None, 'px'), htmlCode=None, helper=None, profile=None):
    """
    Description:
    ------------

    Usage::

      l = rptObj.ui.lists.groups(["AWW", "B"])
      l.add_list(["D", "E"], category="Test")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlList.Groups`

    Related Pages:

    http://designbump.com/create-a-vertical-accordion-menu-using-css3-tutorial/
    http://thecodeplayer.com/walkthrough/vertical-accordion-menu-using-jquery-css3

    Attributes:
    ----------
    :param data:
    :param categories:
    :param color:
    :param width:
    :param height:
    :param htmlCode:
    :param helper:
    :param profile:
    """
    data = data or []
    categories = categories or [""]
    if len(data) > 0:
      if isinstance(data[0], list):
        categories = [""] * len(data)
      else:
        # This object is expecting a list of lists
        data = [data]
    html_obj = html.HtmlList.Groups(self.context.rptObj, data, categories, color, width, height, htmlCode, helper, profile)
    return html_obj

  def tree(self, data=None, color=None, width=('auto', ""), height=(None, 'px'), htmlCode=None, helper=None, options=None, profile=None):
    """
    Description:
    ------------

    Usage::

      data = [{"label": 'test', 'items': [{"label": 'child 1', 'color': 'red'}]}]
      rptObj.ui.lists.tree(data)

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlTrees.Tree`

    ----------
    :param data:
    :param width:
    :param height:
    :param htmlCode:
    :param profile:
    """
    html_tree = html.HtmlTrees.Tree(self.context.rptObj, data or [], color, width, height, htmlCode, helper, options, profile)
    return html_tree

  def dropdown(self, recordSet=None, text="", width=('auto', ""), height=(32, 'px'), htmlCode=None, helper=None, options=None, profile=None):
    """
    Description:
    ------------

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlTrees.DropDown`

    Related Pages:

      http://getbootstrap.com/docs/4.0/components/dropdowns/
      https://www.w3schools.com/bootstrap/tryit.asp?filename=trybs_ref_js_dropdown_multilevel_css&stacked=h
      https://codepen.io/svnt/pen/beEgre

    Attributes:
    ----------
    :param recordSet:
    :param width:
    :param height:
    :param htmlCode:
    :param profile:
    """
    dftl_options = {"width": 90}
    dftl_options.update(options or {})
    html_d = html.HtmlTrees.DropDown(self.context.rptObj, recordSet, text, width, height, htmlCode, helper, dftl_options, profile)
    html_d.style.css.display = 'inline-block'
    return html_d

  def checks(self, data=None, width=('auto', ""), height=(None, 'px'), column=None, htmlCode=None, helper=None, options=None, profile=None):
    """
    Description:
    ------------

    Usage::

      data = [{"label": "python", "value": False}, {"label": "Java", "value": 5}]
      checks = rptObj.ui.lists.checklist(data)

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlList.Checks`

    Related Pages:

    Attributes:
    ----------
    :param data:
    :param width:
    :param height:
    :param column:
    :param htmlCode:
    :param helper:
    :param profile:
    """
    if column is not None:
      values = set()
      for rec in data:
        values.add(rec[column])
      data = sorted(list(values))

    dft_options = {"checked": False}
    if options is not None:
      dft_options.update(options)

    html_list = html.HtmlList.Items(self.context.rptObj, 'check', data or [], width, height, dft_options, htmlCode, profile, helper)
    html_list.css({"list-style": 'none'})
    return html_list

  def badges(self, data=None, width=('auto', ""), height=(None, 'px'), column=None, htmlCode=None, helper=None, options=None, profile=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.lists.badges([{'label': 'Python', 'value': 12}, {'label': 'R', 'value': 3}])

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlList.Badges`

    Related Pages:

      https://www.w3schools.com/bootstrap/bootstrap_list_groups.asp
      https://v4-alpha.getbootstrap.com/components/list-group/

    Attributes:
    ----------
    :param data:
    :param width:
    :param height:
    :param options:
    :param profile:
    """
    if column is not None:
      values = set()
      for rec in data:
        values.add(rec[column])
      data = sorted(list(values))

    dft_options = {"badge": {"background": 'red', 'color': 'white'}}
    if options is not None:
      dft_options.update(options)

    html_list = html.HtmlList.Items(self.context.rptObj, 'badge', data or [], width, height, dft_options, htmlCode, profile, helper)
    html_list.css({"list-style": 'none'})
    return html_list

  def icons(self, data=None, width=('auto', ""), height=(None, 'px'), column=None, htmlCode=None, helper=None, options=None, profile=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.lists.badges([{'label': 'Python', 'value': 12}, {'label': 'R', 'value': 3}])

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlList.Badges`

    Related Pages:

      https://www.w3schools.com/bootstrap/bootstrap_list_groups.asp
      https://v4-alpha.getbootstrap.com/components/list-group/

    Attributes:
    ----------
    :param data:
    :param width:
    :param height:
    :param options:
    :param profile:
    """
    if column is not None:
      values = set()
      for rec in data:
        values.add(rec[column])
      data = sorted(list(values))

    dft_options = {"icon": "fas fa-check"}
    if options is not None:
      dft_options.update(options)

    html_list = html.HtmlList.Items(self.context.rptObj, 'icon', data or [], width, height, dft_options, htmlCode,profile, helper)
    html_list.css({"list-style": 'none'})
    return html_list

  def radios(self, data=None, group_name='group', width=('auto', ""), height=(None, "px"), column=None, htmlCode=None, helper=None, options=None, profile=None):
    """
    Description:
    ------------

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlContainer.Div`
      - :class:`epyk.core.html.HtmlInput.Radio`

    Attributes:
    ----------
    :param data:
    :param group_name:
    :param width:
    :param height:
    :param htmlCode:
    :param helper:
    :param options:
    :param profile:
    """
    if column is not None:
      values = set()
      for rec in data:
        values.add(rec[column])
      data = sorted(list(values))

    html_list = html.HtmlList.Items(self.context.rptObj, 'radio', data or [], width, height, options or {}, htmlCode, profile, helper)
    html_list._jsStyles['group'] = group_name
    html_list.css({"list-style": 'none'})
    return html_list

  def brackets(self, recordSet=None, width=('auto', ""), height=(550, 'px'), options=None, profile=None):
    return html.HtmlList.ListTournaments(self.context.rptObj, recordSet, width, height, options or {}, profile)

  def chips(self, items=None, category='group', placeholder="", width=(100, "%"), height=(60, "px"), htmlCode=None, helper=None, options=None, profile=None):
    """
    Description:
    ------------
    Add a chip (filter) html component

    Usage::

      chips = rptObj.ui.panels.chips()

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlEvent.Filters`

    Related Pages:

			https://www.w3schools.com/howto/howto_css_contact_chips.asp

    Attributes:
    ----------
    :param items: List. Selected items
    :param category: String. The group of the items.
    :param placeholder: String. The input field placeholder
    :param width:
    :param height:
    :param htmlCode:
    :param helper:
    :param profile:
    """
    dflt_options = {"item_css": {"padding": '5px', 'border': '1px solid %s' % self.context.rptObj.theme.success[0], 'border-radius': '5px', 'margin': '2px',
                                 "width": 'auto', 'display': 'inline', 'background': 'white'},
                    'category': category, 'visible': True,
                    'value_css': {'font-size': Defaults.font(3), 'font-weight': 'bold', 'vertical-align': 'bottom'},
                    'category_css': {'display': 'inline', 'margin-right': '2px', 'vertical-align': 'top', 'font-size': Defaults.font(-3)},
                    'icon_css': {'color': self.context.rptObj.theme.success[1], 'margin-left': '5px', 'cursor': 'pointer'}}
    if category == 'group':
      dflt_options['visible'] = False
    if options is not None:
      dflt_options.update(options)
    html_f = html.HtmlEvent.Filters(self.context.rptObj, items or [], width, height, htmlCode, helper, dflt_options, profile)
    html_f.input.attr['placeholder'] = placeholder
    return html_f




from epyk.core.css.styles import GrpCls
from epyk.core.css.styles.attributes import AttrClsContainer

from epyk.core.css import Classes


class ClassDiv(GrpCls.ClassHtml):
  @property
  def css(self):
    """
    Description:
    ------------
    Property to the underlying CSS definition to be added to the style HTML tag of a component

    :rtype: AttrClsContainer.AttrDiv
    """
    if self._css_struct is None:
      self._css_struct = AttrClsContainer.AttrDiv(self.htmlObj)
    return self._css_struct

  @property
  def css_class(self):
    """
    Description:
    ------------
    The internal class used to put a custom Style to this object.
    Only 1 CSS class can be added to an HTML object

    :rtype: Classes.CatalogDiv.CatalogDiv
    """
    if self._css_class is None:
      self._css_class = Classes.CatalogDiv.CatalogDiv(self.htmlObj._report, self.classList['main'], html_id=self.htmlObj.htmlCode).no_border()
    return self._css_class


class ClassModal(GrpCls.ClassHtml):

  @property
  def css(self):
    """
    Description:
    ------------
    Property to the underlying CSS definition to be added to the style HTML tag of a component

    :rtype: AttrClsContainer.AttrModal
    """
    if self._css_struct is None:
      self._css_struct = AttrClsContainer.AttrModal(self.htmlObj)
    return self._css_struct

  @property
  def css_class(self):
    """
    Description:
    ------------
    The internal class used to put a custom Style to this object.
    Only 1 CSS class can be added to an HTML object

    :rtype: Classes.CatalogDiv.CatalogDiv
    """
    if self._css_class is None:
      self._css_class = Classes.CatalogDiv.CatalogDiv(self.htmlObj._report, self.classList['main'], html_id=self.htmlObj.htmlCode).modal()
    return self._css_class


class ClassDrawer(GrpCls.ClassHtml):

  def __init__(self, htmlObj):
    super(ClassDrawer, self).__init__(htmlObj)
    self._css_class_drawer, self._css_class_handle, self._css_class_content, self._css_class_nav = 4 * [None]
    self.classList['main'].add(self.css_class_drawer)
    self.classList['main'].add(self.css_class_handle)
    self.classList['main'].add(self.css_class_content)
    self.classList['main'].add(self.css_class_nav)

  @property
  def css(self):
    """
    Description:
    ------------
    Property to the underlying CSS definition to be added to the style HTML tag of a component

    :rtype: AttrClsContainer.AttrDiv
    """
    if self._css_struct is None:
      self._css_struct = AttrClsContainer.AttrDiv(self.htmlObj)
    return self._css_struct

  @property
  def css_class_drawer(self):
    """
    Description:
    ------------
    The internal class used to put a custom Style to this object.
    Only 1 CSS class can be added to an HTML object

    :rtype: Classes.CatalogDiv.CatalogDrawer
    """
    if self._css_class_drawer is None:
      self._css_class_drawer = Classes.CatalogDiv.CatalogDrawer(self.htmlObj._report, self.classList['main'], html_id=self.htmlObj.htmlCode).drawer()
    return self._css_class_drawer

  @property
  def css_class_nav(self):
    """
    Description:
    ------------
    The internal class used to put a custom Style to this object.
    Only 1 CSS class can be added to an HTML object

    :rtype: Classes.CatalogDiv.CatalogDrawer
    """
    if self._css_class_nav is None:
      self._css_class_nav = Classes.CatalogDiv.CatalogDrawer(self.htmlObj._report, self.classList['main'],
                                                                html_id=self.htmlObj.htmlCode).drawer()
    return self._css_class_nav

  @property
  def css_class_handle(self):
    """
    Description:
    ------------
    The internal class used to put a custom Style to this object.
    Only 1 CSS class can be added to an HTML object

    :rtype: Classes.CatalogDiv.CatalogDrawer
    """
    if self._css_class_handle is None:
      self._css_class_handle = Classes.CatalogDiv.CatalogDrawer(self.htmlObj._report, self.classList['main'],
                                                                html_id=self.htmlObj.htmlCode).handle()
    return self._css_class_handle

  @property
  def css_class_content(self):
    """
    Description:
    ------------
    The internal class used to put a custom Style to this object.
    Only 1 CSS class can be added to an HTML object

    :rtype: Classes.CatalogDiv.CatalogDrawer
    """
    if self._css_class_content is None:
      self._css_class_content = Classes.CatalogDiv.CatalogDrawer(self.htmlObj._report, self.classList['main'],
                                                                html_id=self.htmlObj.htmlCode).content()
    return self._css_class_content





from epyk.core.data.DataClass import DataClass

from epyk.core.html import Html

from epyk.core.js.packages import JsC3
from epyk.core.js import JsUtils
from epyk.core.js.primitives import JsObjects
from epyk.core.js.packages import JsD3


class Chart(Html.Html):
  name = 'C3 Chart'
  requirements = ('c3', )

  def __init__(self, report, width, height, htmlCode, options, profile):
    self.height = height[0]
    super(Chart, self).__init__(report, [], htmlCode=htmlCode, css_attrs={"width": width, "height": height},
                                profile=profile)
    self._d3, self._datasets, self._options, self._data_attrs, self._attrs = None, [], None, {}, {}
    self._options_init = options
    self._options_init['type'] = self._type
    self.style.css.margin_top = 10

  @property
  def chartId(self):
    """
    Description:
    -----------
    Return the Javascript variable of the chart
    """
    return "%s_obj" % self.htmlCode

  @property
  def js(self):
    """
    Description:
    -----------
    JC3 reference API

    Related Pages:

			https://c3js.org/reference.html#api-show

    :return: A Javascript object

    :rtype: JsC3.C3
    """
    if self._js is None:
      self._js = JsC3.C3(self, varName=self.chartId, report=self._report)
    return self._js

  def click(self, jsFncs, profile=False):
    """
    Description:
    -----------
    Set a callback for click event on each data point.

    Related Pages:

			https://c3js.org/reference.html#data-onclick

		Attributes:
    ----------
    :param jsFncs:
    :param profile:
    """
    self.data.onclick(jsFncs, profile)
    return self

  @property
  def d3(self):
    """
    Description:
    -----------

    """
    if self._d3 is None:
      self._d3 = JsD3.D3Select(self._report, selector="d3.select('#%s')" % self.htmlCode, setVar=False)
    return self._d3

  def convert(self, data, options, profile=False):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param data:
    :param options:
    :param profile:
    """
    mod_name = __name__.split(".")[-1]
    constructors = self._report._props.setdefault("js", {}).setdefault("constructors", {})
    constructors[self.builder_name] = "function %s%sConvert(data, options){%s; return result}" % (
      mod_name, self.builder_name, self._js__convertor__)
    if isinstance(data, dict):
      # check if there is no nested HTML components in the data
      tmp_data = ["%s: %s" % (JsUtils.jsConvertData(k, None), JsUtils.jsConvertData(v, None)) for k, v in data.items()]
      js_data = "{%s}" % ",".join(tmp_data)
    else:
      js_data = JsUtils.jsConvertData(data, None)
    dfl_options, js_options = dict(self._options_init), []
    if options is not None:
      dfl_options.update(options)
    for k, v in dfl_options.items():
      if isinstance(v, dict):
        row = ["'%s': %s" % (s_k, JsUtils.jsConvertData(s_v, None)) for s_k, s_v in v.items()]
        js_options.append("'%s': {%s}" % (k, ", ".join(row)))
      else:
        if str(v).strip().startswith("function"):
          js_options.append("%s: %s" % (k, v))
        else:
          js_options.append("%s: %s" % (k, JsUtils.jsConvertData(v, None)))
    return "%s%sConvert(%s, %s)" % (mod_name, self.builder_name, js_data, "{%s}" % ",".join(js_options))

  def build(self, data=None, options=None, profile=False):
    if data:
      return '%(chartId)s.unload(); %(chartId)s.load(%(data)s)' % {'chartId': self.chartId, 'data': self.convert(data, options, profile)}

    return '%s = c3.generate(%s)' % (self.chartId, self.getCtx())

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    return '<div %s></div>' % self.get_attrs(pyClassNames=self.style.get_classes())


class JsTick(DataClass):

  @property
  def format(self):
    return self._attrs["format"]

  @format.setter
  def format(self, val):
    self._attrs["format"] = val


class C3Axis(DataClass):

  @property
  def type(self):
    return self._attrs["type"]

  @type.setter
  def type(self, val):
    self._attrs["type"] = val

  @property
  def show(self):
    return self._attrs["show"]

  @show.setter
  def show(self, val):
    self._attrs["show"] = val

  @property
  def tick(self):
    return self.sub_data("x", JsTick)


class C3Selection(DataClass):
  @property
  def enabled(self):
    """
    Description:
    -----------
    Set data selection enabled.

    If this option is set true, we can select the data points and get/set its state of selection by API (e.g. select, unselect, selected).

    Related Pages:

			https://c3js.org/reference.html#data-selection-enabled
    """
    return self._attrs["enabled"]

  @enabled.setter
  def enabled(self, val):
    self._attrs["enabled"] = val

  @property
  def grouped(self):
    """
    """
    return self._attrs["grouped"]

  @grouped.setter
  def grouped(self, val):
    self._attrs["grouped"] = val

  @property
  def multiple(self):
    """
    """
    return self._attrs["multiple"]

  @multiple.setter
  def multiple(self, val):
    self._attrs["multiple"] = val

  @property
  def draggable(self):
    """
    """
    return self._attrs["draggable"]

  @draggable.setter
  def draggable(self, val):
    self._attrs["draggable"] = val

  @property
  def isselectable(self):
    """
    """
    return self._attrs["isselectable"]

  @isselectable.setter
  def isselectable(self, jsFnc):
    self._attrs["isselectable"] = jsFnc


class JsData(DataClass):
  @property
  def x(self):
    return self._attrs["x"]

  @x.setter
  def x(self, val):
    self._attrs["x"] = val

  @property
  def xs(self): return self._attrs["xs"]

  @xs.setter
  def xs(self, val): self._attrs["xs"] = val

  @property
  def xFormat(self): return self._attrs["xFormat"]

  @xFormat.setter
  def xFormat(self, val): self._attrs["xFormat"] = val

  @property
  def names(self): return self._attrs["names"]

  @names.setter
  def names(self, val): self._attrs["names"] = val

  @property
  def groups(self): return self._attrs["groups"]

  @groups.setter
  def groups(self, val): self._attrs["groups"] = val

  @property
  def axes(self): return self._attrs["axes"]

  @axes.setter
  def axes(self, val): self._attrs["axes"] = val

  @property
  def type(self): return self._attrs["type"]

  @type.setter
  def type(self, val): self._attrs["type"] = val

  def add_type(self, alias, type):
    if "types" not in self._attrs:
      self.types = {}
    self.types[alias] = type
    return self

  @property
  def types(self):
    """
    This setting overwrites data.type setting:
    line, spline, step, area...

    Related Pages:

			https://c3js.org/reference.html#data-types
    """
    return self._attrs["types"]

  @types.setter
  def types(self, val): self._attrs["types"] = val

  @property
  def labels(self): return self._attrs["labels"]

  @labels.setter
  def labels(self, val): self._attrs["labels"] = val

  @property
  def order(self): return self._attrs["order"]

  @order.setter
  def order(self, val): self._attrs["order"] = val

  @property
  def colors(self):
    if "colors" not in self._attrs:
      self._attrs["colors"] = {}
    return self._attrs["colors"]

  @colors.setter
  def colors(self, val): self._attrs["colors"] = val

  @property
  def columns(self):
    if 'columns' not in self._attrs:
      self._attrs["columns"] = []
    return self._attrs["columns"]

  @columns.setter
  def columns(self, val): self._attrs["columns"] = val

  @property
  def hide(self): return self._attrs["hide"]

  @hide.setter
  def hide(self, val): self._attrs["hide"] = val

  @property
  def selection(self):
    return self.sub_data("selection", C3Selection)

  def onclick(self, jsFncs, profile=False):
    self._attrs["onclick"] = JsObjects.JsObject.JsObject("function () { %s }" % JsUtils.jsConvertFncs(jsFncs, toStr=True))


class JsDataEpochs(JsData):

  @property
  def epochs(self):
    """
    """
    return self._attrs["epochs"]

  @epochs.setter
  def epochs(self, val):
    self._attrs["epochs"] = val


class JsScales(DataClass):

  @property
  def rotated(self):
    return self._attrs["rotated"]

  @rotated.setter
  def rotated(self, val):
    self._attrs["rotated"] = val

  @property
  def x(self):
    return self.sub_data("x", C3Axis)

  @property
  def y(self):
    return self.sub_data("y", C3Axis)

  @property
  def y2(self):
    """

    Related Pages:

			https://c3js.org/reference.html#axis-y2-show

    :rtype: C3Axis
    """
    return self.sub_data("y2", C3Axis)


class C3GridLine(DataClass):
  @property
  def value(self):
    return self._attrs["value"]

  @value.setter
  def value(self, val):
    self._attrs["value"] = val

  @property
  def text(self):
    return self._attrs["text"]

  @text.setter
  def text(self, val):
    self._attrs["text"] = val

  @property
  def css_class(self):
    return self._attrs["class"]

  @css_class.setter
  def css_class(self, val):
    self._attrs["class"] = val

  @property
  def position(self):
    return self._attrs["position"]

  @position.setter
  def position(self, val):
    self._attrs["position"] = val


class C3GridAxis(DataClass):

  @property
  def show(self):
    return self._attrs["show"]

  @show.setter
  def show(self, val):
    self._attrs["show"] = val

  @property
  def lines(self):
    return self.sub_data_enum("lines", C3GridLine)


class C3Grid(DataClass):

  @property
  def x(self):
    return self.sub_data("x", C3GridAxis)

  @property
  def y(self):
    return self.sub_data("y", C3GridAxis)


class JsLegend(DataClass):
  pass


class JsTooltip(DataClass):
  pass


class JsSubchart(DataClass):
  pass


class JsZoom(DataClass):
  @property
  def enabled(self):
    return self._attrs["enabled"]

  @enabled.setter
  def enabled(self, val):
    self._attrs["enabled"] = val

  @property
  def type(self):
    return self._attrs["type"]

  @type.setter
  def type(self, val):
    self._attrs["type"] = val

  @property
  def rescale(self):
    return self._attrs["rescale"]

  @rescale.setter
  def rescale(self, val):
    self._attrs["rescale"] = val

  @property
  def extent(self):
    return self._attrs["extent"]

  @extent.setter
  def extent(self, val):
    self._attrs["extent"] = val


class C3Points(DataClass):

  @property
  def show(self):
    return self._attrs["show"]

  @show.setter
  def show(self, val):
    self._attrs["show"] = val

  @property
  def r(self):
    return self._attrs["r"]

  @r.setter
  def r(self, val):
    self._attrs["r"] = val

  @property
  def focus(self):
    return self._attrs["focus"]

  @focus.setter
  def focus(self, val):
    self._attrs["focus"] = {"expand": val, 'enabled': True}

  @property
  def select(self):
    return self._attrs["select"]

  @select.setter
  def select(self, val):
    self._attrs["select"] = val


class ChartLine(Chart):
  _type = 'line'

  def __init__(self, report, width, height, htmlCode, options, profile):
    super(ChartLine, self).__init__(report, width, height, htmlCode, options, profile)
    self._attrs["bindto"] = "#%s" % self.htmlCode

  def labels(self, labels, series_id='x'):
    self.data.x = series_id
    self.data.columns.append([series_id] + labels)

  def add_dataset(self, name, data, type=None):
    self.data.columns.append([name] + data)
    self.data.colors[name] = self._report.theme.colors[len(self.data.colors)]
    if type is None:
      self.data.add_type(name, self._type)
    return self._attrs

  @property
  def _js__convertor__(self):
    return '''
      if(data.python){ 
        result = {'columns': [], type: options.type};
        result['columns'].push(['x'].concat(data.labels));
        data.series.forEach(function(name, i){
          result['columns'].push([name].concat(data.datasets[i]));
        }); 
      } else {
        var temp = {}; var labels = []; var uniqLabels = {};
        options.y_columns.forEach(function(series){temp[series] = {}});
        data.forEach(function(rec){ 
          options.y_columns.forEach(function(name){
            if(rec[name] !== undefined){
              if (!(rec[options.x_column] in uniqLabels)){labels.push(rec[options.x_column]); uniqLabels[rec[options.x_column]] = true};
              temp[name][rec[options.x_column]] = rec[name]}})});
        columns = []; columns.push(['x'].concat(labels));
        options.y_columns.forEach(function(series){
          dataSet = [series];
          labels.forEach(function(x){
            if(temp[series][x] == undefined){dataSet.push(null)} else {dataSet.push(temp[series][x])}}); columns.push(dataSet)});
        var result = {columns: columns}
      }'''

  @property
  def axis(self):
    """

    :rtype: JsScales
    :return:
    """
    if not'axis' in self._attrs:
      self._attrs['axis'] = JsScales(self._report)
    return self._attrs['axis']

  @property
  def point(self):
    """

    :rtype: C3Points
    """
    if not 'point' in self._attrs:
      self._attrs['point'] = C3Points(self._report)
    return self._attrs['point']

  @property
  def zoom(self):
    """

    :rtype: JsZoom
    """
    if not 'zoom' in self._attrs:
      self._attrs['zoom'] = JsZoom(self._report)
    return self._attrs['zoom']

  @property
  def data(self):
    """

    :rtype: JsScales
    """
    if not 'data' in self._attrs:
      self._attrs['data'] = JsData(self._report)
    return self._attrs['data']

  @property
  def grid(self):
    """

    :rtype: JsScales
    :return:
    """
    if not 'grid' in self._attrs:
      self._attrs['grid'] = C3Grid(self._report)
    return self._attrs['grid']

  def getCtx(self):
    str_ctx = "{%s}" % ", ".join(["%s: %s" % (k, JsUtils.jsConvertData(v, None)) for k, v in self._attrs.items()])
    return str_ctx


class ChartSpline(ChartLine):
  _type = 'spline'


class ChartArea(ChartLine):
  _type = 'area'


class ChartBar(ChartLine):
  _type = 'bar'


class ChartScatter(ChartLine):
  _type = 'scatter'

  @property
  def _js__convertor__(self):
    return '''
      if(data.python){ 
        result = {'columns': [], type: options.type};
        result['columns'].push(['x'].concat(data.labels));
        data.series.forEach(function(name, i){
          result['columns'].push([name].concat(data.datasets[i]));
        });
      } else {
        var tempVal = {}; var tempX = {}; var labels = []; 
        options.y_columns.forEach(function(series){tempVal[series] = []; tempX[series +"_x"] = []});
        data.forEach(function(rec){ 
          options.y_columns.forEach(function(name){
            if(rec[name] !== undefined){
              if(!(rec[options.x_column] in tempVal[name])){tempVal[name] = [name, rec[name]]; tempX[name +"_x"] = [name +"_x", rec[options.x_column]]}
              else {tempVal[name].push(rec[name]); tempX[name +"_x"].push(rec[options.x_column])}}})});
        result = {'columns': [], 'xs': {}};
        options.y_columns.forEach(function(series){
          result.columns.push(tempVal[series]); result.columns.push(tempX[series +"_x"]); result.xs[series] = series +"_x"})
      }'''


class ChartPie(ChartLine):
  _type = 'pie'

  @property
  def _js__convertor__(self):
    return '''
      if(data.python){ 
        result = {'columns': [], type: options.type};
        result['columns'].push(['x'].concat(data.labels));
        data.series.forEach(function(name, i){
          result['columns'].push([name].concat(data.datasets[i]));
        }); 
      } else {
        var temp = {}; var labels = {};
        data.forEach(function(rec){ 
          if(!(rec[options.x_column] in temp)){temp[rec[options.x_column]] = {}};
          options.y_columns.forEach(function(name){
            labels[name] = true; 
            if(rec[name] !== undefined){
              if(!(name in temp[rec[options.x_column]])){temp[rec[options.x_column]][name] = rec[name]} else{temp[rec[options.x_column]][name] += rec[name]}}})});
        columns = []; var labels = Object.keys(labels); var count = 0;
        for(var series in temp){
          var values = [count]; count += 1;
          labels.forEach(function(label){
            if(temp[series][label] !== undefined){values.push(temp[series][label])} else{values.push(null)}});
          columns.push(values)};
        var result = {columns: columns}
      }'''

  def labels(self, labels, series_id='x'):
    self._labels = labels

  def add_dataset(self, name, values, type=None):
    for i, value in enumerate(values):
      self.data.columns.append([self._labels[i], value])
      self.data.colors[self._labels[i]] = self._report.theme.colors[i]
      if type is None:
        self.data.add_type(self._labels[i], self._type)
    return self._attrs


class ChartDonut(ChartPie):
  _type = 'donut'


class ChartGauge(ChartPie):
  _type = 'gauge'

  def build(self, data=None, options=None, profile=False):
    if data:
      return '%(chartId)s.load({columns: [["data", %(value)s]]})' % {'chartId': self.chartId, 'value': data}

    return '%s = c3.generate(%s)' % (self.chartId, self.getCtx())

  def add_dataset(self, name, value, type=None):
    self.data.columns.append(["data", value])
    self.data.colors["data"] = self._report.theme.colors[len(self.data.colors)]
    if type is None:
      self.data.add_type("data", self._type)
    return self._attrs


class ChartStanford(ChartPie):
  _type = 'stanford'

  @property
  def data(self):
    """

    :rtype: JsScales
    """
    if not 'data' in self._attrs:
      self._attrs['data'] = JsDataEpochs(self._report)
    return self._attrs['data']

  def epoch(self, series, name):
    self.data.epochs = JsUtils.jsConvertData(str(name), None)
    self.data.columns.append([str(name)] + series)

  def add_dataset(self, name, data, type=None):
    self.data.columns.append([name]+ data)
    if type is None:
      self.data.type = self._type
    return self._attrs



"""
Draft version

"""

from epyk.core import Page
from epyk.core.js import Imports

import sys
import re
import os
import json
import subprocess


def app(path, name=None):
  """
  Description:
  ------------

  Attributes:
  ----------
  :param path:
  """
  return Angular(path, name)


def requirements(report):
  """
  Description:
  ------------
  Get the list of all the packages required in the Angular Application

  Packages can be installed in the app using the command
    > nmp install package1 package2 .....

  Attributes:
  ----------
  :param report: Python object. The report object
  """
  npms = []
  importMng = Imports.ImportManager(online=True, report=report)
  for req in importMng.cleanImports(report.jsImports, Imports.JS_IMPORTS):
    if 'register' in Imports.JS_IMPORTS[req]:
      if 'npm' in Imports.JS_IMPORTS[req]['register']:
        npms.append(Imports.JS_IMPORTS[req]['register']['npm'])
      else:
        print("No npm requirement defined for %s" % req)
    else:
      print("No npm requirement defined for %s" % req)
  return npms


class NGModule(object):

  def __init__(self, ang_app_path):
    self._ang_app_path = ang_app_path

  def class_(self, name):
    """
    Description:
    ------------
    Creates a new generic class definition in the given or default project.

    Related Pages:

      https://angular.io/cli/generate#class-command

    Attributes:
    ----------
    :param name: String. The name of the interface
    """
    subprocess.run('ng generate class %s' % name, shell=True, cwd=self._ang_app_path)

  def component(self, name):
    """
    Description:
    ------------
    Creates a new generic component definition in the given or default project.

    Related Pages:

			https://angular.io/cli/generate#component-command

    Attributes:
    ----------
    :param name: String. The name of the interface
    """
    subprocess.run('ng generate component %s' % name, shell=True, cwd=self._ang_app_path)

  def directive(self, name):
    """
    Description:
    ------------
    Creates a new generic directive definition in the given or default project.

    Related Pages:

			https://angular.io/cli/generate#directive-command

    Attributes:
    ----------
    :param name: String. The name of the interface
    """
    subprocess.run('ng generate directive %s' % name, shell=True, cwd=self._ang_app_path)

  def enum(self, name):
    """
    Description:
    ------------
    Generates a new, generic enum definition for the given or default project.

    Related Pages:

			https://angular.io/cli/generate#enum-command

    Attributes:
    ----------
    :param name: String. The name of the interface
    """
    subprocess.run('ng generate enum %s' % name, shell=True, cwd=self._ang_app_path)

  def guard(self, name):
    """
    Description:
    ------------
    Generates a new, generic route guard definition in the given or default project.

    Related Pages:

      https://angular.io/cli/generate#guard-command

    Attributes:
    ----------
    :param name: String. The name of the new route guard.
    """
    subprocess.run('ng generate guard %s' % name, shell=True, cwd=self._ang_app_path)

  def interceptor(self, name):
    """
    Description:
    ------------
    Creates a new, generic interceptor definition in the given or default project.

    Related Pages:

      https://angular.io/cli/generate#interceptor-command

    Attributes:
    ----------
    :param name: String. The name of the interceptor.
    """
    subprocess.run('ng generate interceptor %s' % name, shell=True, cwd=self._ang_app_path)

  def interface(self, name, type):
    """
    Description:
    ------------
    Creates a new generic interface definition in the given or default project.

    Related Pages:

      https://angular.io/cli/generate#interface-command

    Attributes:
    ----------
    :param name: String. The name of the interface
    :param type: String. Adds a developer-defined type to the filename, in the format "name.type.ts".
    """
    subprocess.run('ng generate interface %s %s' % (name, type), shell=True, cwd=self._ang_app_path)

  def library(self, name, type):
    """
    Description:
    ------------
    Creates a new generic library project in the current workspace.

    Related Pages:

      https://angular.io/cli/generate#library-command

    Attributes:
    ----------
    :param name: String. The name of the interface
    """
    subprocess.run('ng generate library %s %s' % (name, type), shell=True, cwd=self._ang_app_path)

  def module(self, name, type):
    """
    Description:
    ------------
    Creates a new generic NgModule definition in the given or default project

    Related Pages:

      https://angular.io/cli/generate#library-command

    Attributes:
    ----------
    :param name: String. The name of the interface
    """
    subprocess.run('ng generate module %s %s' % (name, type), shell=True, cwd=self._ang_app_path)

  def service(self, name, type):
    """
    Description:
    ------------
    Creates a new, generic service definition in the given or default project.

    Related Pages:

      https://angular.io/cli/generate#library-command

    Attributes:
    ----------
    :param name: String. The name of the interface
    """
    subprocess.run('ng generate service %s %s' % (name, type), shell=True, cwd=self._ang_app_path)


class NG(object):
  def __init__(self, app_path, app_name=None):
    self._app_path, self._app_name = app_path, app_name

  def e2e(self, app_name=None):
    """
    Description:
    ------------
    Builds and serves an Angular app, then runs end-to-end tests using Protractor.

    Related Pages:

      https://angular.io/cli/e2e

    Attributes:
    ----------
    :param app_name:
    """
    app_name = app_name or self._app_name
    if app_name is None:
      raise Exception("An Angular aplication name is required!")

    subprocess.run('ng e2e %s' % app_name, shell=True, cwd=os.path.join(self._app_path, self._app_name))

  def lint(self, app_name=None):
    """
    Description:
    ------------
    Builds and serves an Angular app, then runs end-to-end tests using Protractor.

    Related Pages:

			https://angular.io/cli/lint

		Attributes:
    ----------
    :param app_name:
    """
    app_name = app_name or self._app_name
    if app_name is None:
      raise Exception("An Angular aplication name is required!")

    subprocess.run('ng lint %s' % app_name, shell=True, cwd=os.path.join(self._app_path, self._app_name))

  def new(self, name, path=None):
    """
    Description:
    ------------
    Builds and serves an Angular app, then runs end-to-end tests using Protractor.

    Related Pages:

			https://angular.io/cli/new

		Attributes:
    ----------
    :param name:
    :param path:
    """
    if path is not None:
      subprocess.run('ng new %s --directory %s' % (name, path), shell=True, cwd=self._app_path)
    else:
      subprocess.run('ng new %s' % name, shell=True, cwd=self._app_path)
    print('ng new %s' % name)

  def doc(self, keyword):
    """
    Description:
    ------------
    Opens the official Angular documentation (angular.io) in a browser, and searches for a given keyword.

    Related Pages:

			https://angular.io/cli

		Attributes:
    ----------
    :param keyword:
    """
    subprocess.run('ng doc %s' % keyword, shell=True, cwd=self._app_path)

  def add(self, package):
    """
    Description:
    ------------

    Related Pages:

			https://angular.io/cli

		Attributes:
    ----------
    :param package:
    """
    subprocess.run('ng add %s' % package, shell=True, cwd=os.path.join(self._app_path, self._app_name))
    print("%s packages installed" % package)

  def analytics(self):
    """
    https://angular.io/cli
    """
    pass

  def help(self, options=None):
    """
    Description:
    ------------
    Lists available commands and their short descriptions.

    Related Pages:

			https://angular.io/cli

		Attributes:
    ----------
    :param app_name:
    """
    if options is None:
      subprocess.run('ng help', shell=True, cwd=self._app_path)
    else:
      subprocess.run('ng help %s' % options, shell=True, cwd=os.path.join(self._app_path, self._app_name))

  def test(self, app_name=None):
    """
    Description:
    ------------

    Related Pages:

			https://angular.io/cli

    Attributes:
    ----------
    :param app_name:
    """
    app_name = app_name or self._app_name
    if app_name is None:
      raise Exception("An Angular aplication name is required!")

    subprocess.run('ng test %s' % app_name, shell=True, cwd=os.path.join(self._app_path, app_name))

  def build(self, app_name=None):
    """
    Description:
    ------------
    Compiles an Angular app into an output directory named dist/ at the given output path. Must be executed from within a workspace directory.

    Related Pages:

			https://angular.io/cli

    Attributes:
    ----------
    :param app_name:
    """
    app_name = app_name or self._app_name
    if app_name is None:
      raise Exception("An Angular aplication name is required!")

    subprocess.run('ng build %s' % app_name, shell=True, cwd=os.path.join(self._app_path, app_name))

  def version(self):
    """
    Description:
    ------------
    Builds and serves an Angular app, then runs end-to-end tests using Protractor.
    """
    subprocess.run('ng version', shell=True, cwd=os.path.join(self._app_path, self._app_name))

  def serve(self):
    """
    Description:
    ------------
    Builds and serves an Angular app, then runs end-to-end tests using Protractor.

    Related Pages:

			https://angular.io/cli/serve
    """
    subprocess.run('ng serve --open', shell=True, cwd=os.path.join(self._app_path, self._app_name))

  @property
  def create(self):
    """
    Description:
    ------------
    Shortcut to the various generate entry points in the Angular Framework
    """
    return NGModule(os.path.join(self._app_path, self._app_name))

  def generate(self, schematic, name):
    """
    Description:
    ------------
    Generates and/or modifies files based on a schematic.

    Related Pages:

			https://angular.io/cli/generate

    Attributes:
    ----------
    :param schematic:
    :param name:
    """
    subprocess.run('ng generate %s %s' % (schematic, name), shell=True, cwd=os.path.join(self._app_path, self._app_name))


class RouteModule(object):

  def __init__(self, app_path, app_name, file_name=None):
    self._app_path, self._app_name = app_path, app_name
    self.file_name = file_name or 'app-routing.module.ts'
    self.modules, self.routes, self.ng_modules = {}, {}, None

    # parse the Angular route module
    imported_module_pattern = re.compile("import { (.*) } from '(.*)';")
    imported_route_pattern = re.compile("{ path: '(.*)', component: (.*) }")

    self.ngModule = "@NgModule"
    with open(os.path.join(self._app_path, app_name, 'src', 'app', 'app-routing.module.ts')) as f:
      content = f.read()
      split_content = content.split("@NgModule")
      for module_def in imported_module_pattern.findall(split_content[0]):
        self.modules[module_def[0]] = module_def[1]

      for alias, component in imported_route_pattern.findall(split_content[0]):
        self.routes[component] = alias
      self.ngModule = "%s%s" % (self.ngModule, split_content[1])

  def add(self, component, alias, path):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param component:
    :param alias:
    :param path:
    """
    if self.ng_modules is None:
      self.ng_modules = NgModules(self._app_path, self._app_name)
      self.ng_modules.modules[component] = path
    self.modules[component] = path
    self.routes[component] = alias

  def export(self, file_name=None, target_path=None):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param file_name: String. Optional. The filename
    """
    file_name = 'app-routing.module_new.ts' or self.file_name
    if target_path is None:
      target_path = []
    target_path.append(file_name)
    with open(os.path.join(self._app_path, self._app_name, *target_path), "w") as f:
      for k, v in self.modules.items():
        f.write("import { %s } from '%s';\n" % (k, v))
      f.write("\n\n")
      f.write("const routes: Routes = [\n")
      for k, v in self.routes.items():
        f.write("    { path: '%s', component: %s },\n" % (v, k))
      f.write("]\n\n")
      f.write("\n")
      f.write(self.ngModule)
    self.ng_modules.export(file_name)


class NgModules(object):

  def __init__(self, app_path, app_name, file_name=None):
    self._app_path, self._app_name = app_path, app_name
    self.file_name = file_name or 'app.module.ts'
    self.modules, self.imports, self.declarations, self.providers, self.bootstrap = {}, [], [], [], []

    # parse the Angular route module
    imported_module_pattern = re.compile("import { (.*) } from '(.*)';")
    with open(os.path.join(self._app_path, app_name, 'src', 'app', self.file_name)) as f:
      content = f.read()
      for module_def in imported_module_pattern.findall(content):
        self.modules[module_def[0]] = module_def[1]

      matches = re.search(r"\@NgModule\((.*)?\)", content, re.DOTALL)
      json_stuff = re.sub(r"([a-z_\-\.A-Z0-9]+)", r'"\1"', matches.group(1))
      json_stuff = re.sub(r",[\W]*?([\}\]])", r"\n\1", json_stuff)
      json_dict = json.loads(json_stuff)
      self.declarations = json_dict['declarations']
      self.providers = json_dict['providers']
      self.imports = json_dict['imports']
      self.bootstrap = json_dict['bootstrap']

    # Add mandatory core modules
    self.add_import("HttpClientModule", '@angular/common/http')

  def add(self, component, path):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param component: String. The component name
    :param path: String. The component path
    """
    self.modules[component] = path

  def add_import(self, component, path):
    self.modules[component] = path

    if component not in self.imports:
      self.imports.append(component)

  def export(self, file_name=None, target_path=None):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param file_name: String. Optional. The filename
    """
    file_name = 'app.module_new.ts' or self.file_name
    if target_path is None:
      target_path = []
    target_path.append(file_name)
    with open(os.path.join(self._app_path, self._app_name, *target_path), "w") as f:
      for k, v in self.modules.items():
        f.write("import { %s } from '%s';\n" % (k, v))
      f.write("\n\n")
      f.write("@NgModule({\n")
      for name, vars in [("declarations", self.declarations), ('imports', self.imports), ('providers', self.providers), ('bootstrap', self.bootstrap)]:
        f.write("  %s: [\n" % name)
        for d in vars:
          f.write("    %s,\n" % d)
        f.write("  ],\n")
      f.write("})\n\n")
      f.write("export class AppModule { }")


class ComponentSpec(object):

  def __init__(self, app_path, app_name, alias, name):
    self.imports, self.vars = {}, {}
    self._app_path, self._app_name, self.__path = app_path, app_name, 'apps'
    self.alias, self.name = alias, name
    self.__comp_structure = {}

  def export(self, path=None, target_path=None):
    """
    Description:
    ------------
    Export the spec of the component

    TODO: make this generation more flexible

    Attributes:
    ----------
    :param path:
    :param target_path: for example ['src', 'app']
    """
    self.__path = path or self.__path
    if target_path is None:
      target_path = []
    target_path.append(self.__path)
    module_path = os.path.join(self._app_path, self._app_name, *target_path)
    if not os.path.exists(module_path):
      os.makedirs(module_path)

    with open(os.path.join(module_path, "app.%s.component.spec.ts" % self.alias), "w") as f:
      f.write('''
import { TestBed, async } from '@angular/core/testing';
import { RouterTestingModule } from '@angular/router/testing';
import { %(name)s } from './app.component';

describe('%(name)s', () => {
  beforeEach(async(() => {
    TestBed.configureTestingModule({
      imports: [
        RouterTestingModule
      ],
      declarations: [
        %(name)s
      ],
    }).compileComponents();
  }));

  it('should create the app', () => {
    const fixture = TestBed.createComponent(%(name)s);
    const app = fixture.componentInstance;
    expect(app).toBeTruthy();
  });

  it(`should have as title '%(alias)s'`, () => {
    const fixture = TestBed.createComponent(%(name)s);
    const app = fixture.componentInstance;
    expect(app.title).toEqual('%(alias)s');
  });

  it('should render title', () => {
    const fixture = TestBed.createComponent(%(name)s);
    fixture.detectChanges();
    const compiled = fixture.nativeElement;
    expect(compiled.querySelector('.content span').textContent).toContain('%(alias)s app is running!');
  });
});
''' % {"path": module_path, 'name': self.name, 'alias': self.alias})


class Components(object):

  def __init__(self, app, count=0, report=None):
    self._app, self.count_comp, self._report = app, count, report

  def chart(self, vars, alias=None):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param vars:
    :param alias:
    """
    map_var_names = {}
    alias = alias or self.count_comp
    for k, v in vars.items():
      map_var_names[k] = "%s%s" % (k, alias)
      self._app.add_var(map_var_names[k], v)
    map_var_names["ref"] = "chart%s" % alias
    self._app.comps["chart%s" % alias] = "chart%s" % alias
    self.count_comp += 1
    self._app.htmls.append('''
<div class="container">
		<pr-chart [values]='%(values)s' [labels]='%(labels)s' type='{{ %(type)s }}' #%(ref)s></pr-chart>
</div>''' % map_var_names)
    return "chart%s" % alias

  def table(self, vars, alias=None):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param vars:
    :param alias:
    """
    map_var_names = {}
    alias = alias or self.count_comp
    for k, v in vars.items():
      map_var_names[k] = "%s%s" % (k, alias)
      self._app.add_var(map_var_names[k], v)
    map_var_names["ref"] = "table%s" % alias
    self._app.comps["table%s" % alias] = "table%s" % alias
    self.count_comp += 1
    self._app.htmls.append('''<pr-grid #%(ref)s></pr-grid>''' % map_var_names)
    return "table%s" % alias

  def input(self):
    self._app.comps["input"] = "input"
    self._app.htmls.append('''<input type='text' #input />''')

  def button(self, value):
    return self._report.ui.button(value)


class App(object):

  def __init__(self, app_path, app_name, alias, name, report=None):
    self.imports = {'Component': '@angular/core', 'HttpClient': '@angular/common/http',
                    'HttpHeaders': '@angular/common/http', 'Injectable': '@angular/core',
                    'ViewChild': '@angular/core'}
    self.vars, self.__map_var_names, self._report = {}, {}, report
    self._app_path, self._app_name = app_path, app_name
    self.alias, self.__path, self.className, self.__components = alias, 'apps', name, None
    self.__comp_structure, self.htmls, self.__fncs, self.__injectable_prop = {}, [], {}, {'providedIn': 'root'}
    self.spec = ComponentSpec(app_path, app_name,  alias, name)
    self.comps = {}

  def add_var(self, name, value=None):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param name:
    :param value:
    """
    if value is not None:
      if 'constructor' not in self.__comp_structure:
        self.__comp_structure['constructor'] = []
      self.__comp_structure['constructor'].append("this.%s = %s" % (name, json.dumps(value)))
    self.vars[name] = value

  def add_fnc(self, name, fncs):
    self.__fncs[name] = fncs

  def add_imports(self, name, path):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param name:
    :param path:
    """
    pass

  @property
  def components(self):
    """
    Description:
    ------------

    """
    if self.__components is None:
      self.__components = Components(self, report=self._report)
    return self.__components

  def constructor(self):
    pass

  def ngOnInit(self):
    pass

  def ngAfterViewInit(self):
    return

  @property
  def name(self):
    """
    Description:
    ------------
    Return the prefix of the component module (without any extension)
    """
    return "app.%s.component" % self.alias

  def http(self, end_point, jsFncs):
    return "this.httpClient.post('http://127.0.0.1:5000/%s', {}).subscribe((data)=>{ %s });" % (end_point, ";".join(jsFncs))

  @property
  def path(self):
    """
    Description:
    ------------
    Return the full path of the component modules
    """
    return os.path.join("./", self.__path, self.name).replace("\\", "/")

  def export(self, path=None, target_path=None):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param path:
    :param target_path: for example ['src', 'app']
    """
    self.__path = path or self.__path
    if target_path is None:
      target_path = []
    target_path.append(self.__path)
    module_path = os.path.join(self._app_path, self._app_name, *target_path)
    if not os.path.exists(module_path):
      os.makedirs(module_path)

    self.spec.export(path=path, target_path=None)
    for component in self._report.components.values():
      if component.options.managed:
        print(component.properties())

    with open(os.path.join(module_path, "%s.ts" % self.name), "w") as f:
      for comp, path in self.imports.items():
        f.write("import { %s } from '%s';\n" % (comp, path))
      f.write("\n")
      # All the applicatops need this as they will interact with a Flask backend
      f.write("@Injectable({\n")
      for k , v in self.__injectable_prop.items():
        f.write(" %s: %s\n" % (k, json.dumps(v)))
      f.write("})\n")
      f.write("\n")
      f.write("@Component({\n")
      f.write("  selector: 'app-%s',\n" % self.alias)
      f.write("  templateUrl: '%s',\n" % "./%s.html" % self.name)
      f.write("  styleUrls: ['%s']\n" % "./%s.css" % self.name)
      f.write("})\n")
      f.write("\n")
      f.write("export class %s {\n" % self.className)
      f.write("  // Component link section\n")
      for k, v in self.comps.items():
        f.write("  @ViewChild('%s') %s: any;;\n" % (k, k))

      f.write("  // Variable section\n")
      for var in self.vars.keys():
        f.write("  %s;\n" % var)
      f.write("\n")
      if 'constructor' in self.__comp_structure:
        f.write("  constructor(private httpClient: HttpClient){\n")
        f.write("    %s" % "; ".join(self.__comp_structure['constructor']))
        f.write("  }\n")
      f.write("\n")
      for name, fnc_def in self.__fncs.items():
        f.write("  %s(){ %s } \n" % (name, ";".join(fnc_def)))
      f.write("}\n")
    with open(os.path.join(module_path, "%s.html" % self.name), "w") as f:
      f.write("".join(self.htmls))

    with open(os.path.join(module_path, "%s.css" %  self.name), "w") as f:
      f.write("")


class Angular(object):

  def __init__(self, app_path, name=None):
    self._app_path, self._app_name = app_path, name
    self.__route, self.__ng_modules = None, None
    self.__page = None

  def create(self, node_path):
    """
    Description:
    ------------

    """

  def npm(self, packages):
    """
    Description:
    ------------
    Use npm install on a package.
    Can be done directly on the nodeJs app using the command line:
      npm install package1 package2 .....

    Attributes:
    ----------
    :param packages: List. The list of packages to install retrieved from requirements()
    """
    subprocess.run('npm install %s --save' % " ".join(packages), shell=True, cwd=self._app_path)
    print("%s packages installed" % len(packages))

  def ls(self):
    """
    Description:
    ------------
    Search the registry for packages matching terms
    """
    subprocess.run('npm ls', shell=True, cwd=self._app_path)

  def docs(self, package):
    """
    Description:
    ------------
    Display the README.md / documentation / npmjs.org page of a give library

    :param package: String. The package alias
    """
    subprocess.run('npm docs %s' % package, shell=True, cwd=self._app_path)

  def update(self, packages):
    """
    Description:
    ------------
    Update all the packages listed to the latest version (specified by the tag config). Also install missing packages

    Attributes:
    ----------
    :param packages:
    """
    subprocess.run('npm update %s' % " ".join(packages), shell=True, cwd=self._app_path)
    print("%s packages updated" % len(packages))

  def uninstall(self, packages):
    """
    Description:
    ------------
    Uninstall a package, completely removing everything npm installed on its behalf

    Attributes:
    ----------
    :param packages:
    """
    subprocess.run('npm uninstall %s' % " ".join(packages), shell=True, cwd=self._app_path)
    print("%s packages uninstalled" % len(packages))

  def install(self):
    """
    Description:
    ------------
    Install or update Angular on the defined path
    """
    subprocess.run('npm install -g @angular/cli', shell=True, cwd=self._app_path)
    print("Angular CLI installed")

  def ng(self, app_name=None):
    """
    Description:
    ------------
    Angular specific command lines

    Related Pages:

			https://angular.io/cli/

    Attributes:
    ----------
    :param app_name: String. The angular application name
    """
    app_name = app_name or self._app_name
    return NG(self._app_path, app_name)

  def page(self, selector=None, name=None, report=None, auto_route=False):
    """
    Description:
    ------------
    Create a specific Application as a component in the Angular framework.

    Unlike a basic component, the application will be routed to be accessed directly.

    Description:
    ------------
    :param report: Object. A report object
    :param selector: String. The url route for this report in the Angular app
    :param name: String. The component classname in the Angular framework
    """
    if name is None:
      script = os.path.split(sys._getframe().f_back.f_code.co_filename)[1][:-3]
      name = "".join([s.capitalize() for s in script.split("_")])
      if selector is None:
        selector = script.replace("_", "-")
    report = report or Page.Report()
    report.framework("ANGULAR")
    self.__page = App(self._app_path, self._app_name, selector, name, report=report)
    if auto_route:
      self.route.add(self.__page.className, self.__page.alias, self.__page.path)
    return self.__page

  def ng_modules(self, app_name=None, file_name=None):
    """
    Description:
    ------------
    Read the file app.module.ts

    :param app_name: String. Optinal. THe Angular application name

    :rtype: NgModules
    """
    if self.__ng_modules is None:
      if self.__route is not None and self.__route.ng_modules is not None:
        self.__ng_modules = self.__route.ng_modules
      else:
        self.__ng_modules = NgModules(self._app_path, app_name or self._app_name, file_name)
    return self.__ng_modules

  @property
  def route(self, app_name=None, file_name=None):
    """
    Description:
    ------------

    Read the file app-routing.module.ts from the Angular app

    Attributes:
    ----------
    :param app_name: String. Optinal. THe Angular application name
    :param file_name: String. Optinal.

    :rtype: RouteModule
    """
    if self.__route is None:
      self.__route = RouteModule(self._app_path, app_name or self._app_name, file_name)
    return self.__route

  def publish(self, app_name=None, target_path=None):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param app_name:
    :param target_path: List  for example ['src', 'app']
    """
    if self.__page is not None:
      self.__page.export(target_path=target_path)
    if self.__route is not None:
      self.route.export()




from epyk.core.html import Html
from epyk.core.js.objects import JsCanvas


class Canvas(Html.Html):
  name = 'Canvas'

  def __init__(self, report, width, height):
    super(Canvas, self).__init__(report, "", css_attrs={"width": width, "height": height})
    self.__ctx = None

  @property
  def dom(self):
    """

    :rtype: JsCanvas.Canvas
    """
    if self._dom is None:
      self._dom = JsCanvas.Canvas(self, report=self._report)
    return self._dom

  @property
  def ctx(self):
    """

    :rtype: JsCanvas.Context2D
    """
    if self.__ctx is None:
      self._dom = JsCanvas.Canvas(self, report=self._report)
      self.__ctx = self._dom.getContext2D
    return self.__ctx

  def __str__(self):
    return "<canvas %s>Your browser does not support the HTML5 canvas tag.</canvas>" % (self.get_attrs(pyClassNames=self.style.get_classes()))





from epyk.core.css.styles.classes import CssStyle


class CssAgHead(CssStyle.Style):
  classname = "ag-header-cell-label"

  def customize(self):
    self.css({'color': self.rptObj.theme.greys[0], 'padding': '0 0 0 5px', 'background-color': self.rptObj.theme.greys[-1]})


class CssAgOddRow(CssStyle.Style):
  classname = 'ag-row-odd'

  def customize(self):
    self.css({'color': self.rptObj.theme.greys[-1], 'background-color': self.rptObj.theme.greys[0]}, important=True)


class CssAgEvenRow(CssStyle.Style):
  classname = 'ag-row-even'

  def customize(self):
    self.css({'color': self.rptObj.theme.greys[-1], 'background-color': self.rptObj.theme.colors[1]}, important=True)


class CssAgCellFocus(CssStyle.Style):
  classname = 'ag-cell-focus'

  def customize(self):
    self.css({'border': '1px solid %s' % self.rptObj.theme.success[1]}, important=True)


class CssAgCell(CssStyle.Style):
  classname = 'ag-cell'

  def customize(self):
    self.css({'line-height': '25px', 'padding': '0 0 0 5px'}, important=True)




from epyk.core.css.styles.classes import CssStyle


class CssDropDownSubMenu(CssStyle.Style):
  _attrs = {'display': 'none', 'border': '1px solid rgba(0,0,0,.15)', 'border-radius': '4px',
            'box-shadow': '0 6px 12px rgba(0,0,0,.175)', '-webkit-box-shadow': '0 6px 12px rgba(0,0,0,.175)'}
  classname = "menu .submenu"


class CssDropDownMenuHLi(CssStyle.Style):
  _attrs = {'display': 'block'}
  classname = "menu li.dropdown:hover > .submenu"


class CssDropDownMenuHoverAAfter(CssStyle.Style):
  _attrs = {'display': 'block'}
  classname = "menu ul.submenu > li.dropdown:hover > .submenu"


class CssDropDownSubMenuPullLeft(CssStyle.Style):
  _attrs = {'position': 'absolute', 'top': '0px', 'list-style-type': 'none', 'white-space': 'nowrap', 'margin': 0, 'padding': 0} # 'left': '0px',
  classname = "menu .submenu .submenu"


class CssDropDownMenu(CssStyle.Style):
  _attrs = {'position': 'absolute', 'top': '0px', 'list-style-type': 'none', 'white-space': 'nowrap', 'margin': 0, 'padding': 0} # 'left': '0px',
  classname = "menu ul.submenu"


class CssDropDownAfterMenu(CssStyle.Style):
  _attrs = {'display': 'block'}
  classname = "menu li a"

  def customize(self):
    self.css({'color': self.rptObj.theme.greys[-1], 'background': self.rptObj.theme.greys[0]})


class CssDropDownCaret(CssStyle.Style):
  _attrs = {'display': 'inline-block', "width": 0, "height": 0, 'margin-left': '2px', 'vertical-align': 'middle',
            'border-top': '4px dashed', 'border-right': '4px solid transparent', 'border-left': '4px solid transparent'}
  classname = "caret"




from epyk.core.html import Html

from epyk.core.js import expr
from epyk.core.css import Selector

from epyk.core.html import Defaults
from epyk.core.html.options import OptPanel
from epyk.core.css.styles import GrpClsContainer
from epyk.core.js.html import JsHtmlStepper


class Drawer(Html.Html):
  name = 'Drawer'

  def __init__(self, report, width, height, options, helper, profile):
    super(Drawer, self).__init__(report, None, css_attrs={"width": width, "height": height})
    self.add_helper(helper, css={"line-height": '%spx' % Defaults.LINE_HEIGHT})
    self.__options = OptPanel.OptionDrawer(self, options)
    self.style.css.position = 'relative'

    self.panels = report.ui.div()
    self.panels.options.managed = False
    self.panels.style.css.padding_right = 10
    self.panels.attr['name'] = 'drawer_panels'

    self.handle = report.ui.div()
    self.handle.style.clear_all()
    self.handle.style.css.cursor = 'pointer'

    self.handle.options.managed = False
    self.handle.attr['name'] = 'drawer_handle'

    self.drawers = report.ui.div()
    self.drawers.style.clear_all()
    self.drawers.style.css.overflow_y = 'auto'
    self.drawers.options.managed = False
    self.drawers.attr['name'] = 'drawer_content'

  @property
  def dom(self):
    """
    Description:
    ------------
    Property to get the common dom features

    :rtype: JsHtmlStepper.Drawer
    """
    if self._dom is None:
      self._dom = JsHtmlStepper.Drawer(self, report=self._report)
    return self._dom

  @property
  def options(self):
    """
    Description:
    ------------
    Property to set all the possible object for a drawer

    :rtype: OptPanel.OptionDrawer
    """
    return self.__options

  def add_panel(self, link, container, display=False):
    """
    Description:
    ------------
    Add panel to the drawer object.

    Attributes:
    ----------
    :param link: String | HTML object. The value in the drawer
    :param container: HTML object. The component to be displayed
    :param display: String. The CSS Display property
    """
    if not hasattr(link, 'options'):
      link = self._report.ui.div(link)
      link.style.css.padding = "0 5px"
      link.options.managed = False
    if not hasattr(container, 'options'):
      container = self._report.ui.div(container)
    container.style.css.display = display
    container.options.managed = False
    self.panels += container
    self.drawers += link

  @property
  def style(self):
    """
    Description:
    ------------
    Get the CSS Style of the object

    :rtype: GrpClsContainer.ClassDrawer
    """
    if self._styleObj is None:
      self._styleObj = GrpClsContainer.ClassDrawer(self)
    return self._styleObj

  def set_handle(self, component):
    """
    Description:
    ------------
    Set the handle used to trigger the open / close events
    """
    self.handle = self._report.ui.div()
    self.handle.style.clear_all()
    if self.options.side == 'left':
      component.click([self.drawers.dom.toggle_transition("margin-right", "0px", "-%s" % self.options.width)])
    else:
      component.click([self.drawers.dom.toggle_transition("margin-left", "0px", "-%s" % self.options.width)])

  def __str__(self):
    self.handle.style.css.float = self.options.side
    if self.options.side == 'left':
      self.drawers.style.css.width = self.options.width
      self.drawers.style.css.margin_right = "-%s" % self.options.width
      self.handle.click([self.drawers.dom.toggle_transition("margin-right", "0px", "-%s" % self.options.width)])
    else:
      self.drawers.style.css.width = self.options.width
      self.drawers.style.css.margin_left = "-%s" % self.options.width
      self.handle.click([self.drawers.dom.toggle_transition("margin-left", "0px", "-%s" % self.options.width)])
    position = {"left": 'right', 'right': 'left'}
    return '''
      <div %(attr)s>
        %(panels)s
        <div name='drawer' style='clear:both;%(side)s:0;overflow-y:hidden'>
          %(handle)s%(drawer)s
        </div>
      </div>''' % {'attr': self.get_attrs(pyClassNames=self.style.get_classes()), 'htmlCode': self.htmlCode,
                   'drawer': self.drawers.html(), 'handle': self.handle.html(), 'panels': self.panels.html(), 'side': position[self.options.side]}


class DrawerMulti(Html.Html):
  name = 'Multi Drawers'

  def __init__(self, report, component, width, height, options, helper, profile):
    super(DrawerMulti, self).__init__(report, None, css_attrs={"width": width, "height": height})
    self.add_helper(helper, css={"line-height": '%spx' % Defaults.LINE_HEIGHT})
    self.__options = OptPanel.OptionDrawer(self, options)
    self.style.css.position = 'relative'

    self.panels = component
    self.panels.options.managed = False
    self.panels.style.css.display = "inline-block"
    self.panels.style.css.width = "auto"
    self.panels.style.css.padding_right = 10

    self.handle = report.ui.div()
    self.handle.style.clear_all()
    self.handle.style.css.z_index = 10
    self.handle.style.css.position = 'relative'

    self.handle.options.managed = False
    self.handle.attr['name'] = 'drawer_handle'

    self.drawers = report.ui.div()
    self.drawers.style.clear_all()
    self.drawers.style.css.overflow_y = 'auto'
    self.drawers.options.managed = False
    self.drawers.attr['name'] = 'drawer_content'

  @property
  def dom(self):
    """
    Description:
    ------------
    Property to get the common dom features

    :rtype: JsHtmlStepper.Drawer
    """
    if self._dom is None:
      self._dom = JsHtmlStepper.Drawer(self, report=self._report)
    return self._dom

  @property
  def options(self):
    """
    Description:
    ------------
    Property to set all the possible object for a drawer

    :rtype: OptPanel.OptionDrawer
    """
    return self.__options

  def add_drawer(self, link, container):
    """
    Description:
    ------------
    Add panel to the drawer object.

    Attributes:
    ----------
    :param link: String | HTML object. The value in the drawer
    :param container: HTML object. The component to be displayed
    :param display: String. The CSS Display property
    """
    if not hasattr(link, 'options'):
      link = self._report.ui.div(link)
    link.style.css.padding = "5px 0"
    link.style.css.color = self._report.theme.colors[0]
    link.style.css.margin = "0 2px"
    link.style.css.cursor = "pointer"
    link.options.managed = False
    if not hasattr(container, 'options'):
      container = self._report.ui.div(container)
      container.style.css.padding = 5
    container.options.managed = False
    link.style.css.writing_mode = "vertical-rl"
    link.style.css.text_orientation = "mixed"
    self.handle += link
    self.drawers += container
    link.click([
      self._report.js.querySelectorAll(Selector.Selector(self.drawers).with_child_element("div").excluding(container)).css({"display": 'none'}),
      expr.if_(self.panels.dom.getAttribute("data-panel") == container.htmlCode, [
        self.drawers.dom.toggle_transition("margin-right" if self.options.side == 'left' else "margin-left", "-%s" % self.options.width, "0px"),
        container.dom.css({"display": 'none'}),
        self.panels.dom.setAttribute("data-panel", '')])
      .else_([
        expr.if_(self._report.js.querySelector(Selector.Selector(self.drawers)).css("margin-left") != "0px", [
          self.drawers.dom.toggle_transition("margin-right" if self.options.side == 'left' else "margin-left", "0px", "-%s" % self.options.width),
        ]),
        self.panels.dom.setAttribute("data-panel", container.htmlCode),
        container.dom.css({'display': 'block'})
      ])
    ])

  @property
  def style(self):
    """
    Description:
    ------------
    Get the CSS Style of the object

    :rtype: GrpClsContainer.ClassDrawer
    """
    if self._styleObj is None:
      self._styleObj = GrpClsContainer.ClassDrawer(self)
    return self._styleObj

  def __str__(self):
    self.handle.style.css.float = self.options.side
    if self.options.side == 'left':
      self.drawers.style.css.width = self.options.width
      self.drawers.style.css.margin_right = "-%s" % self.options.width
    else:
      self.drawers.style.css.width = self.options.width
      self.drawers.style.css.margin_left = "-%s" % self.options.width
    return '''
      <div %(attr)s>
        %(panels)s
        <div name='drawer' style='clear:both;%(side)s:0;overflow-y:hidden'>
          %(drawer)s
        </div>
        %(handle)s
      </div>''' % {'attr': self.get_attrs(pyClassNames=self.style.get_classes()), 'htmlCode': self.htmlCode,
                   'drawer': self.drawers.html(), 'handle': self.handle.html(), 'panels': self.panels.html(), 'side': self.options.side}





"""
Module in charge of the PivotTable library

Related Pages:

		https://pivottable.js.org/examples/
"""

from epyk.core.html import Html
from epyk.core.js import JsUtils
from epyk.core.html.options import OptTable
from epyk.core.js.packages import JsQuery

# The list of CSS classes
from epyk.core.css.styles import GrpClsTable


class PivotTable(Html.Html):
  requirements = ('pivot', )
  name = 'Pivot Table'

  def __init__(self, report, recordSet, rows, cols, width, height, htmlCode, helper, options, profile):
    super(PivotTable, self).__init__(report, recordSet, htmlCode=htmlCode, profile=profile, css_attrs={"width": width, "height": height})
    # Add the extra HTML components
    self.add_helper(helper)
    self.__options = OptTable.OptionsPivot(self, options)
    # to add all the columns in the table if nothing defined
    self._jsStyles.update({'cols': cols or [], 'rows': rows or []})
    self.style.css.display = 'block'

  @property
  def style(self):
    if self._styleObj is None:
      self._styleObj = GrpClsTable.Pivot(self)
    return self._styleObj

  @property
  def options(self):
    """
    Description:
    ------------
    Pivot Table options

    :rtype: OptTable.OptionsPivot
    """
    return self.__options

  @property
  def aggregators(self):
    return PivotAggregator(self._report, self._jsStyles)

  @property
  def renderers(self):
    return PivotRenderer(self._report, self._jsStyles)

  @property
  def _js__builder__(self):
    return '''
      if (options.showUI){%(jqId)s.pivotUI(data, options)}
      else {%(jqId)s.pivot(data, options)}
      ''' % {"jqId": JsQuery.decorate_var("htmlObj", convert_var=False)}

  def sub_total(self):
    """

    :return:
    """
    self._report.jsImports.add('pivot-sub-total')
    self.options.dataClass = "$.pivotUtilities.SubtotalPivotData"
    self.options.renderers = "$.pivotUtilities.subtotal_renderers"
    self.options.rendererName = 'Table With Subtotal'

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    return '<div %(strAttr)s></div>%(helper)s' % {'strAttr': self.get_attrs(pyClassNames=self.style.get_classes()), "helper": self.helper}


class PivotUITable(PivotTable):
  __reqJs, __reqCss = ["pivot"], ["pivot"]

  def __init__(self, report, recordSet, rows, cols, width, height, htmlCode, helper, options, profile):
    super(PivotUITable, self).__init__(report, recordSet, rows, cols, width, height, htmlCode, helper, options, profile)
    self.__options = OptTable.OptionsPivotUI(self, options)
    # to add all the columns in the table if nothing defined
    self._jsStyles.update({'cols': cols or [], 'rows': rows or []})

  @property
  def options(self):
    """
    Description:
    ------------
    Pivot Table options

    :rtype: OptTable.OptionsPivotUI
    """
    return self.__options

  @property
  def _js__builder__(self):
    return '''
        %(jqId)s.pivotUI(data, options)
        ''' % {"jqId": JsQuery.decorate_var("htmlObj", convert_var=False)}


class PivotAggregator(object):

  def __init__(self, report, options):
    self.report, self.options = report, options

  def sumOverSum(self, cola):
    """
    Description:
    ------------

    :param cola:
    """
    cola = JsUtils.jsConvertData(cola, None)
    self.options['aggregator'] = '$.pivotUtilities.aggregators["Sum over Sum"](%s)' % cola
    self.options['aggregatorName'] = "sumOverSum"

  def count(self):
    """
    Description:
    ------------

    """
    self.options['aggregator'] = '$.pivotUtilities.aggregators["Count"]()'
    self.options['aggregatorName'] = "count"

  def sum(self, col1):
    """
    Description:
    ------------

    """
    col1 = JsUtils.jsConvertData(col1, None)
    self.options['aggregator'] = '$.pivotUtilities.aggregators["Sum"]([%s])' % col1
    self.options['aggregatorName'] = "sum"

  def max(self, col1):
    """
    Description:
    ------------

    """
    self.singleFactorFormulas(col1, "= Math.max(this.tmpVal, col1)")
    self.options['aggregatorName'] = "Max"

  def min(self, col1):
    """
    Description:
    ------------

    """
    self.singleFactorFormulas(col1, "= Math.min(this.tmpVal, col1)")
    self.options['aggregatorName'] = "Min"

  def absSum(self, col1):
    """
    Description:
    ------------

    """
    self.singleFactorFormulas(col1, "+= Math.abs(col1)")
    self.options['aggregatorName'] = "sum (abs)"

  def singleFactorFormulas(self, col1, formula):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param col1:
    :param formula:
    """
    col1 = JsUtils.jsConvertData(col1, None)
    self.options['aggregator'] = '''
      function(keyAgg) { 
        return function(data, rowKey, colKey) {
          return {
            tmpVal: 0, numInputs: 1,
            push: function(record){
              const col1 = record[keyAgg]; this.tmpVal %s; return this.tmpVal},
            value: function() { return this.tmpVal; },
            format: function(x) { return x; },
          }}}(%s)''' % (formula, col1)
    self.options['aggregatorName'] = "diff Abs Agg"

  def twoFactorFormulas(self, col1, col2, formula):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param col1:
    :param col2:
    :param formula:
    """
    col1 = JsUtils.jsConvertData(col1, None)
    col2 = JsUtils.jsConvertData(col2, None)
    self.options['aggregator'] = '''
      function(keyAgg, key2Agg) { 
        return function(data, rowKey, colKey) {
          return {
            tmpVal: 0, numInputs: 2,
            push: function(record){
              const col1 = record[keyAgg]; const col2 = record[key2Agg]; this.tmpVal %s; return this.tmpVal},
            value: function() { return this.tmpVal; },
            format: function(x) { return x; },
          };
        };
      }(%s, %s)''' % (formula, col1, col2)
    self.options['aggregatorName'] = "diff Abs Agg"

  def diffAbsolute(self, col1, col2, formula="+= col1 - col2"):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param col1:
    :param col2:
    :param formula:
    """
    self.twoFactorFormulas(col1, col2, formula)
    self.options['aggregatorName'] = "diff Abs Agg"

  def custom(self, name, js_def):
    """

    https://github.com/nicolaskruchten/pivottable/wiki/Aggregators

    Attributes:
    ----------
    :param name:
    :param js_def:
    """
    self.options['aggregator'] = js_def
    self.options['aggregatorName'] = name


class PivotRendererC3(object):
  def __init__(self, report, options):
    self._report, self.options = report, options

  def bar(self):
    """
    Description:
    ------------
    Horizontal bar chart from C3

    Related Pages:

			https://pivottable.js.org/examples/c3.html

    """
    self._report.jsImports.add('pivot-c3')
    self.options['renderers'] = "$.extend($.pivotUtilities.renderers, $.pivotUtilities.c3_renderers)"
    self.options['rendererName'] = "Bar Chart"

  def scatter(self):
    """
    Description:
    ------------
    Scatter chart from C3

    Related Pages:

			https://pivottable.js.org/examples/c3.html
    """
    self._report.jsImports.add('pivot-c3')
    self.options['renderers'] = "$.extend($.pivotUtilities.renderers, $.pivotUtilities.c3_renderers)"
    self.options['rendererName'] = "Area Chart"

  def area(self):
    """
    Description:
    ------------
    Area chart from C3

    Related Pages:

			https://pivottable.js.org/examples/c3.html
    """
    self._report.jsImports.add('pivot-c3')
    self.options['renderers'] = "$.extend($.pivotUtilities.renderers, $.pivotUtilities.c3_renderers)"
    self.options['rendererName'] = "Area Chart"

  def line(self):
    """
    Description:
    ------------
    Line chart from C3

    Related Pages:

			https://pivottable.js.org/examples/c3.html
    """
    self._report.jsImports.add('pivot-c3')
    self.options['renderers'] = "$.extend($.pivotUtilities.renderers, $.pivotUtilities.c3_renderers)"
    self.options['rendererName'] = "Line Chart"

  def hbar(self):
    """
    Description:
    ------------
    Stacked bar chart from C3

    Related Pages:

			https://pivottable.js.org/examples/c3.html
    """
    self._report.jsImports.add('pivot-c3')
    self.options['renderers'] = "$.extend($.pivotUtilities.renderers, $.pivotUtilities.c3_renderers)"
    self.options['rendererName'] = "Horizontal Stacked Bar Chart"

  def stacked(self):
    """
    Description:
    ------------
    Stacked bar chart from C3

    Related Pages:

			https://pivottable.js.org/examples/c3.html
    """
    self._report.jsImports.add('pivot-c3')
    self.options['renderers'] = "$.extend($.pivotUtilities.renderers, $.pivotUtilities.c3_renderers)"
    self.options['rendererName'] = "Stacked Bar Chart"


class PivotRendererPlotly(object):

  def __init__(self, report, options):
    self._report, self.options = report, options

  def pies(self):
    """
    Description:
    ------------
    Multiple Pies charts from Plotly

    Related Pages:

			https://pivottable.js.org/examples/plotly.html
    """
    self._report.jsImports.add('pivot-plotly')
    self.options['renderers'] = "$.extend($.pivotUtilities.renderers, $.pivotUtilities.plotly_renderers)"
    self.options['rendererName'] = "Multiple Pie Chart"

  def area(self):
    """
    Description:
    ------------
    Area chart from Plotly

    Related Pages:

			https://pivottable.js.org/examples/c3.html
    """
    self._report.jsImports.add('pivot-plotly')
    self.options['renderers'] = "$.extend($.pivotUtilities.renderers, $.pivotUtilities.plotly_renderers)"
    self.options['rendererName'] = "Area Chart"

  def scatter(self):
    """
    Description:
    ------------
    Scatter chart from Plotly

    Related Pages:

			https://pivottable.js.org/examples/c3.html
    """
    self._report.jsImports.add('pivot-plotly')
    self.options['renderers'] = "$.extend($.pivotUtilities.renderers, $.pivotUtilities.plotly_renderers)"
    self.options['rendererName'] = "Scatter Chart"

  def line(self):
    """
    Description:
    ------------
    Line chart from Plotly

    Related Pages:

			https://pivottable.js.org/examples/c3.html
    """
    self._report.jsImports.add('pivot-plotly')
    self.options['renderers'] = "$.extend($.pivotUtilities.renderers, $.pivotUtilities.plotly_renderers)"
    self.options['rendererName'] = "Line Chart"

  def bar(self):
    """
    Description:
    ------------
    Bar chart from Plotly

    Related Pages:

			https://pivottable.js.org/examples/c3.html
    """
    self._report.jsImports.add('pivot-plotly')
    self.options['renderers'] = "$.extend($.pivotUtilities.renderers, $.pivotUtilities.plotly_renderers)"
    self.options['rendererName'] = "Bar Chart"

  def hbar(self):
    """
    Description:
    ------------
    Horizontal Bar chart from Plotly

    Related Pages:

			https://pivottable.js.org/examples/c3.html
    """
    self._report.jsImports.add('pivot-plotly')
    self.options['renderers'] = "$.extend($.pivotUtilities.renderers, $.pivotUtilities.plotly_renderers)"
    self.options['rendererName'] = "Horizontal Bar Chart"


class PivotRenderer(object):

  def __init__(self, report, options):
    self._report, self.options = report, options

  def table(self):
    self.options['renderer'] = '$.pivotUtilities.renderers["table"]'

  @property
  def plotly(self):
    """
    Property to use the Plotly special renderers
    """
    return PivotRendererPlotly(self._report, self.options)

  @property
  def c3(self):
    """
    Property to use the C3 special renderers
    """
    return PivotRendererC3(self._report, self.options)

  def treemap(self):
    """
    https://pivottable.js.org/examples/plotly.html

    """
    self._report.jsImports.add('pivot-d3')
    self.options['renderers'] = "$.pivotUtilities.d3_renderers"
    self.options['rendererName'] = "Treemap"

  def heatmap(self):
    self.options['renderer'] = '$.pivotUtilities.renderers["Heatmap"]'

  def bars(self):
    self.options['renderer'] = '$.pivotUtilities.renderers["Table Barchart"]'

  def custom(self, name, js_def):
    """

    https://github.com/nicolaskruchten/pivottable/wiki/Renderers

    :param name:
    :param js_def:
    """
    pass



from epyk.core.js.packages import packageImport
from epyk.core.data.DataClass import DataGroup


class ExtsFormattors(DataGroup):

  @packageImport('tabulator-icons')
  def icon(self, css=None, tags=None, **kwargs):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param css:
    :param tags:
    :param kwargs:
    """
    self._attrs["formatter"] = 'icon'
    self._attrs["formatterParams"] = {}
    if css is not None:
      self._attrs["css"] = css
    if tags is not None:
      self._attrs["tags"] = tags
    for k, v in kwargs.items():
      self._attrs["formatterParams"][k] = v
    return self

  @packageImport('tabulator-icons')
  def icon_pivot(self, iconMapping, pivot=None, cssMapping=None, tags=None, **kwargs):
    """
    Description:
    -----------
    Set an icon in the cell based on a lookup table based on another value in the row
    Default will take the cell value as icon classname

    Attributes:
    ----------
    :param pivot: String. The column field in the row
    :param iconMapping: Dictionary. A Icon classname mapping
    :param cssMapping: Dictionary. A CSS mapping for the icons containers
    :param tags: Optional. A dictionary with the different Dom tags to be addeed
    :param kwargs:
    """
    self._attrs["formatter"] = 'iconMapPivot'
    self._attrs["formatterParams"] = {'pivot': pivot or self._attrs["field"], 'iconMapping': iconMapping}
    if cssMapping is not None:
      self._attrs["cssMapping"] = cssMapping
    if tags is not None:
      self._attrs["tags"] = tags
    for k, v in kwargs.items():
      self._attrs["formatterParams"][k] = v
    return self

  @packageImport('tabulator-inputs')
  def password(self, css=None, **kwargs):
    """
    Description:
    -----------
    Change the content of the cell to ****

    Attributes:
    ----------
    :param css: Dictionary. The CSS attributes for the cell (Optional)
    :param kwargs:
    """
    self._attrs["formatter"] = 'password'
    formatParams = {}
    if css is not None:
      formatParams['css'] = css
    for k, v in kwargs.items():
      formatParams[k] = v
    return self

  @packageImport('tabulator-numbers')
  def label_thresholds(self, thresholds, labels, css=None, **kwargs):
    """
    Description:
    -----------
    Set a label based on a list of values

    Attributes:
    ----------
    :param thresholds: List. The different values to compare to deduce the category
    :param labels: List. The resulting category
    :param css: Dictionary. The CSS attributes for the cell (Optional)
    :param kwargs:
    """  #
    self._attrs["formatter"] = 'labelThresholds'
    formatParams = {'thresholds': thresholds, 'labels': labels}
    if css is not None:
      formatParams['css'] = css
    if kwargs:
      self._attrs["formatterParams"].update(kwargs)
    return self

  @packageImport('tabulator-numbers')
  def label_thresholds_pivot(self, pivot, thresholds, labels, css=None, **kwargs):
    """
    Description:
    -----------
    Set a label based on a list of values from another column

    Attributes:
    ----------
    :param pivot: String. The column name to use to get the data to lookup from te row
    :param thresholds: List. The different values to compare to deduce the category
    :param labels: List. The resulting category
    :param css: Dictionary. The CSS attributes for the cell (Optional)
    :param kwargs:
    """  #
    self._attrs["formatter"] = 'flagThresholdsPivot'
    formatParams = {'thresholds': thresholds, 'labels': labels, 'pivot': pivot}
    if css is not None:
      formatParams['css'] = css
    if kwargs:
      self._attrs["formatterParams"].update(kwargs)
    return self

  @packageImport('tabulator-numbers')
  def number(self, decimal=".", thousand=",", precision=0, symbol="", format="%v", css=None, **kwargs):
    """

    :param decimal: String. decimal point separator default "."
    :param thousand: String. thousands separator default ","
    :param precision: Integer. decimal places default 0
    :param symbol: default currency symbol is ''
    :param format:
    :param css: Dictionary. The CSS attributes for the cell (Optional)
    :param kwargs:
    """  #
    self._attrs["formatter"] = 'numbers'
    self._attrs["formatterParams"] = {k: v for k, v in locals().items() if k != 'self' and v is not None}
    self._attrs["formatterParams"].update(self._attrs["formatterParams"].pop('kwargs'))
    return self

  @packageImport('tabulator-numbers')
  def number_format(self, decimal=".", thousand=",", precision=0, symbol="", format="%v", colors=None,
                              threshold=0, css=None, **kwargs):
    """
    Description:
    -----------
    
    Attributes:
    ----------
    :param decimal: String. decimal point separator default "."
    :param thousand: String. thousands separator default ","
    :param precision: Integer. decimal places default 0
    :param symbol: default currency symbol is ''
    :param format: String. "%s%v" controls output: %s = symbol, %v = value/number (can be object: see below)
    :param css: Dictionary. The CSS attributes for the cell (Optional)
    :param kwargs:
    """  #
    if colors is None:
      colors = [self._report.theme.danger[1], self._report.theme.success[1]]
    self._attrs["formatter"] = 'numbersFormat'
    self._attrs["formatterParams"] = {k: v for k, v in locals().items() if k != 'self' and v is not None}
    self._attrs["formatterParams"].update(self._attrs["formatterParams"].pop('kwargs'))
    return self

  @packageImport('tabulator-numbers')
  def number_difference(self, decimal=None, thousand=None, precision=None, symbol=None, format=None,
                                  colors=None, threshold=0, css=None, **kwargs):
    """
    Description:
    -----------
    
    Attributes:
    ----------
    :param decimal: String. decimal point separator default "."
    :param thousand: String. thousands separator default ","
    :param precision: Integer. decimal places default 0
    :param symbol: default currency symbol is ''
    :param format: String. "%s%v" controls output: %s = symbol, %v = value/number (can be object: see below)
    :param colors: List. Color before and after the threshold (default red and green according to the theme)
    :param threshold: Integer. The threshold number
    :param css: Dictionary. The CSS attributes for the cell (Optional)
    :param kwargs:
    """  #
    if colors is None:
      colors = [self._report.theme.danger[1], self._report.theme.success[1]]
    self._attrs["formatter"] = 'numbersDifference'
    self._attrs["formatterParams"] = {k: v for k, v in locals().items() if k != 'self' and v is not None}
    self._attrs["formatterParams"].update(self._attrs["formatterParams"].pop('kwargs'))
    return self

  @packageImport('tabulator-numbers')
  def number_thresholds(self, thresholds, css, **kwargs):
    """
    Description:
    -----------
    
    Attributes:
    ----------
    :param thresholds:
    :param css:
    :param kwargs:
    """  #
    self._attrs["formatter"] = 'numbersThreshold'
    self._attrs["formatterParams"] = {'thresholds': thresholds, 'css': css}
    self._attrs["formatterParams"] = {k: v for k, v in locals().items() if k != 'self' and v is not None}
    self._attrs["formatterParams"].update(self._attrs["formatterParams"].pop('kwargs'))
    return self

  @packageImport('tabulator-numbers')
  def number_thresholds_pivot(self, pivot, thresholds, css, **kwargs):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param pivot:
    :param thresholds:
    :param css:
    :param kwargs:
    """
    self._attrs["formatter"] = 'numbersThresholdPivot'
    self._attrs["formatterParams"] = {'thresholds': thresholds, 'css': css, 'pivot': pivot}
    self._attrs["formatterParams"] = {k: v for k, v in locals().items() if k != 'self' and v is not None}
    self._attrs["formatterParams"].update(self._attrs["formatterParams"].pop('kwargs'))
    return self

  @packageImport('tabulator-drop')
  def drag_and_drop(self, row_delimiter="\n", col_delimiter="\t", css=None, **kwargs):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param row_delimiter: String
    :param col_delimiter: String
    :param css:
    :param kwargs:
    """
    self._attrs["formatter"] = 'dragAndDrop'
    self._attrs["formatterParams"] = {'css': css, 'rowDelimiter': row_delimiter, 'colDelimiter': col_delimiter}
    self._attrs["formatterParams"].update({k: v for k, v in locals().items() if k != 'self' and v is not None})
    self._attrs["formatterParams"].update(self._attrs["formatterParams"].pop('kwargs'))
    return self

  @packageImport('tabulator-numbers')
  def intensity(self, steps, colors, intensity, css=None, **kwargs):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param steps:
    :param colors:
    :param intensity: String, The column used to deduce the intensity. Default the cell value
    :param css:
    :param kwargs:
    """
    self._attrs["formatter"] = 'numbersIntensity'
    self._attrs["formatterParams"] = {'steps': steps, 'colors': colors, 'intensity': intensity}
    if css is not None:
      self._attrs["formatterParams"]['css'] = css
    self._attrs["formatterParams"].update({k: v for k, v in locals().items() if k != 'self' and v is not None})
    self._attrs["formatterParams"].update(self._attrs["formatterParams"].pop('kwargs'))
    return self

  @packageImport('tabulator-numbers')
  def quality(self, steps, colors, intensity, quality, css=None, **kwargs):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param steps:
    :param colors:
    :param intensity:
    :param quality:
    :param css:
    :param kwargs:
    """
    self._attrs["formatter"] = 'numbersIntensity'
    self._attrs["formatterParams"] = {'steps': steps, 'colors': colors, 'intensity': intensity, 'quality': quality}
    if css is not None:
      self._attrs["formatterParams"]['css'] = css
    self._attrs["formatterParams"] = {k: v for k, v in locals().items() if k != 'self' and v is not None}
    self._attrs["formatterParams"].update(self._attrs["formatterParams"].pop('kwargs'))
    return self

  @packageImport('tabulator-inputs')
  def lookup_pivot(self, lookups, pivot, css=None, **kwargs):
    """
    Description:
    -----------
    Set a label based on a list of values

    Attributes:
    ----------
    :param lookups:
    :param pivot:
    :param css: Dictionary. The CSS attributes for the cell (Optional)
    """
    self._attrs["formatter"] = 'lookupPivot'
    formatParams = {'lookups': lookups, "pivot": pivot}
    if css is not None:
      formatParams['css'] = css
    for k, v in kwargs.items():
      formatParams[k] = v
    self._attrs['formatterParams'] = formatParams
    return self

  def buttonTick(self, **kwargs):
    """
    Description:
    -----------
    The buttonTick formater displays a tick icon on each row (for use as a button)

    Related Pages:
http://tabulator.info/docs/4.1/format
    """
    self._attrs["formatter"] = 'buttonTick'
    if kwargs:
      self._attrs["editorParams"] = kwargs
    return self

  def buttonCross(self, **kwargs):
    """
    Description:
    -----------
    The buttonCross formater displays a cross icon on each row (for use as a button)

    Related Pages:
http://tabulator.info/docs/4.1/format
    """
    self._attrs["formatter"] = 'buttonCross'
    if kwargs:
      self._attrs["editorParams"] = kwargs
    return self

  def rownum(self, **kwargs):
    """
    Description:
    -----------
    The rownum formatter shows an incrementing row number for each row as it is displayed

    Related Pages:
http://tabulator.info/docs/4.1/format
    """
    self._attrs["formatter"] = 'rownum'
    if kwargs:
      self._attrs["editorParams"] = kwargs
    return self

  def handle(self, **kwargs):
    """
    Description:
    -----------
    The handle formatter fills the cell with hamburger bars, to be used as a row handle

    Related Pages:
http://tabulator.info/docs/4.1/format
    """
    self._attrs["formatter"] = 'handle'
    if kwargs:
      self._attrs["editorParams"] = kwargs
    return self

  @packageImport('tabulator-inputs')
  def style(self, css=None, valField=None, cssField=None, **kwargs):
    """
    Description:
    -----------

    Related Pages:
http://tabulator.info/docs/4.1/format

    Attributes:
    ----------
    :param css: Dictionary for the css style to apply
    :param cssField: Dictionary for the css style to apply
    :param valField:
    """
    if css is None and cssField is None:
      raise Exception("Both CSS and CSSField cannot be empty")

    self._attrs["formatter"] = 'cssStyle'
    self._attrs['formatterParams'] = {}
    if css is not None:
      self._attrs['formatterParams']['css'] = css
    if cssField is not None:
      self._attrs['formatterParams']['cssField'] = cssField
      self._attrs['formatterParams']['valField'] = valField
    if kwargs:
      self._attrs["formatterParams"].update(kwargs)
    return self

  @packageImport('tabulator-inputs')
  def style_pivot(self, cssMapping, pivot=None, **kwargs):
    self._attrs["formatter"] = 'cssStylePivot'
    self._attrs['formatterParams'] = {'cssMapping': cssMapping}
    if pivot is not None:
      self._attrs['formatterParams']['pivot'] = pivot
    if kwargs:
      self._attrs["formatterParams"].update(kwargs)
    return self

  def custom(self, formatter, formatterParams, moduleAlias):
    """

    :param formatter:
    :param formatterParams:
    :param moduleAlias:
    """
    self._report.jsImports.add(moduleAlias)
    self._attrs["formatter"] = formatter
    self._attrs['formatterParams'] = formatterParams
    return self




from epyk.core.html import graph


class C3(object):
  def __init__(self, context):
    self.parent = context
    self.chartFamily = "C3"

  def line(self, record=None, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """
    Description:
    ------------
    Display a line chart from C3

    Related Pages:

      https://c3js.org/reference.html#line-connectNull

    Attributes:
    ----------
    :param record: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the record
    :param profile:
    :param width: Tuple. The width of the component in the page, default (100, '%')
    :param height: Tuple. The height of the component in the page, default (330, "px")
    :param options:
    :param htmlCode:
    """
    options = options or {}
    options.update({'y_columns': y_columns or [], 'x_column': x_axis})
    data = self.parent.context.rptObj.data.c3.y(record, y_columns, x_axis)
    line_chart = graph.GraphC3.ChartLine(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    line_chart.labels(data['labels'])
    for i, d in enumerate(data['datasets']):
      line_chart.add_dataset(data['series'][i], d)
    return line_chart

  def spline(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """
    Description:
    ------------
    Display a spline line chart from C3

    Related Pages:

      https://c3js.org/samples/chart_spline.html

    Attributes:
    ----------
    :param record: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the record
    :param profile:
    :param width: Tuple. The width of the component in the page, default (100, '%')
    :param height: Tuple. The height of the component in the page, default (330, "px")
    :param options:
    :param htmlCode:
    """
    options = options or {}
    options.update({'y_columns': y_columns, 'x_column': x_axis})
    data = self.parent.context.rptObj.data.c3.y(record, y_columns, x_axis)
    line_chart = graph.GraphC3.ChartSpline(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    line_chart.labels(data['labels'])
    for i, d in enumerate(data['datasets']):
      line_chart.add_dataset(data['series'][i], d)
    return line_chart

  def step(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """
    Description:
    ------------
    Display a step line chart from C3

    Related Pages:

      https://c3js.org/samples/chart_step.html

    Attributes:
    ----------
    :param record: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the recor
    :param profile:
    :param width: Tuple. The width of the component in the page, default (100, '%')
    :param height: Tuple. The height of the component in the page, default (330, "px")
    :param options:
    :param htmlCode:
    """
    options = options or {}
    options.update({'y_columns': y_columns, 'x_column': x_axis})
    data = self.parent.context.rptObj.data.c3.y(record, y_columns, x_axis)
    line_chart = graph.GraphC3.ChartSpline(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    line_chart._type = 'step'
    line_chart.labels(data['labels'])
    for i, d in enumerate(data['datasets']):
      line_chart.add_dataset(data['series'][i], d)
    return line_chart

  def area(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """
    Description:
    ------------
    Display a area line chart from C3

    Related Pages:

      https://c3js.org/samples/chart_step.html

    Attributes:
    ----------
    :param record: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the recor
    :param profile:
    :param width: Tuple. The width of the component in the page, default (100, '%')
    :param height: Tuple. The height of the component in the page, default (330, "px")
    :param options:
    :param htmlCode:
    """
    options = options or {}
    options.update({'y_columns': y_columns, 'x_column': x_axis})
    data = self.parent.context.rptObj.data.c3.y(record, y_columns, x_axis)
    line_chart = graph.GraphC3.ChartArea(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    line_chart.labels(data['labels'])
    for i, d in enumerate(data['datasets']):
      line_chart.add_dataset(data['series'][i], d)
    return line_chart

  def area_step(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """
    Description:
    ------------
    Display a area step line chart from C3

    Related Pages:

      https://c3js.org/samples/chart_step.html

    Attributes:
    ----------
    :param record: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the recor
    :param profile:
    :param width: Tuple. The width of the component in the page, default (100, '%')
    :param height: Tuple. The height of the component in the page, default (330, "px")
    :param options:
    :param htmlCode:
    """
    options = options or {}
    options.update({'y_columns': y_columns, 'x_column': x_axis})
    data = self.parent.context.rptObj.data.c3.y(record, y_columns, x_axis)
    line_chart = graph.GraphC3.ChartArea(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    line_chart._type = "area-step"
    line_chart.labels(data['labels'])
    for i, d in enumerate(data['datasets']):
      line_chart.add_dataset(data['series'][i], d)
    return line_chart

  def timeseries(self, record, y_columns=None, x_axis=None, profile=None, options=None, width=(100, "%"), height=(330, "px"), htmlCode=None):
    """
    Description:
    ------------
    Display a timeseries chart from C3

    Related Pages:

      https://c3js.org/samples/timeseries.html

    Attributes:
    ----------
    :param record: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the recor
    :param profile:
    :param options:
    :param width: Tuple. The width of the component in the page, default (100, '%')
    :param height: Tuple. The height of the component in the page, default (330, "px")
    :param htmlCode:
    """
    line = self.line(record, y_columns, x_axis, profile, width, height, options, htmlCode)
    line.axis.x.type = "timeseries"
    line.axis.x.tick.format = "%Y-%m-%d"
    return line

  def bar(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """
    Description:
    ------------
    Display a bar chart from C3

    Related Pages:

      https://c3js.org/samples/chart_bar.html

    Attributes:
    ----------
    :param record: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the recor
    :param profile:
    :param width: Tuple. The width of the component in the page, default (100, '%')
    :param height: Tuple. The height of the component in the page, default (330, "px")
    :param options:
    :param htmlCode:
    """
    options = options or {}
    options.update({'y_columns': y_columns, 'x_column': x_axis})
    data = self.parent.context.rptObj.data.c3.y(record, y_columns, x_axis)
    line_chart = graph.GraphC3.ChartBar(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    line_chart.labels(data['labels'])
    for i, d in enumerate(data['datasets']):
      line_chart.add_dataset(data['series'][i], d)
    return line_chart

  def hbar(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """
    Description:
    ------------
    Display a horizontal bar chart from C3

    Related Pages:

			https://c3js.org/samples/axes_rotated.html

    Attributes:
    ----------
    :param record: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the recor
    :param profile:
    :param width: Tuple. The width of the component in the page, default (100, '%')
    :param height: Tuple. The height of the component in the page, default (330, "px")
    :param options:
    :param htmlCode:
    """
    h_bar = self.bar(record, y_columns, x_axis, profile, width, height, options, htmlCode)
    h_bar.axis.rotated = True
    return h_bar

  def scatter(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """
    Description:
    ------------
    Display a Scatter chart from C3

    Related Pages:

			https://c3js.org/samples/axes_rotated.html

    Attributes:
    ----------
    :param record: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the recor
    :param profile:
    :param width: Tuple. The width of the component in the page, default (100, '%')
    :param height: Tuple. The height of the component in the page, default (330, "px")
    :param options:
    :param htmlCode:
    """
    options = options or {}
    options.update({'y_columns': y_columns, 'x_column': x_axis})
    data = self.parent.context.rptObj.data.c3.y(record, y_columns, x_axis)
    line_chart = graph.GraphC3.ChartScatter(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    line_chart.labels(data['labels'])
    for i, d in enumerate(data['datasets']):
      line_chart.add_dataset(data['series'][i], d)
    return line_chart

  def pie(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """
    Description:
    ------------
    Display a pie chart from C3

    Related Pages:

			https://c3js.org/samples/chart_pie.html
			https://c3js.org/reference.html#pie-label-show

    Attributes:
    ----------
    :param record: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the recor
    :param profile:
    :param width: Tuple. The width of the component in the page, default (100, '%')
    :param height: Tuple. The height of the component in the page, default (330, "px")
    :param options:
    :param htmlCode:
    """
    options = options or {}
    options.update({'y_columns': y_columns, 'x_column': x_axis})
    data = self.parent.context.rptObj.data.c3.y(record, y_columns, x_axis)
    pie_chart = graph.GraphC3.ChartPie(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    pie_chart.labels(data['labels'])
    pie_chart.add_dataset(data['series'][0], data['labels'])
    return pie_chart

  def donut(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """
    Description:
    ------------
    Display a donut chart from C3

    Related Pages:

      https://c3js.org/samples/chart_donut.html

    Attributes:
    ----------
    :param record: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the recor
    :param profile:
    :param width: Tuple. The width of the component in the page, default (100, '%')
    :param height: Tuple. The height of the component in the page, default (330, "px")
    :param options:
    :param htmlCode:
    """
    options = options or {}
    options.update({'y_columns': y_columns, 'x_column': x_axis})
    data = self.parent.context.rptObj.data.c3.y(record, y_columns, x_axis)
    pie_chart = graph.GraphC3.ChartDonut(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    pie_chart.labels(data['labels'])
    pie_chart.add_dataset(data['series'][0], data['labels'])
    return pie_chart

  def gauge(self, value, text="", profile=None, options=None, width=(100, "%"), height=(330, "px"), htmlCode=None):
    """
    Description:
    ------------
    Display a gauge chart from C3

    Related Pages:

        https://c3js.org/samples/chart_gauge.html

    Attributes:
    ----------
    :param value:
    :param text:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    """
    g_chart = graph.GraphC3.ChartGauge(self.parent.context.rptObj, width, height, htmlCode, options or {}, profile)
    g_chart.add_dataset(text, value)
    return g_chart

  def stanford(self, record, y_columns=None, x_axis=None, epoch_col=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """

    :param record:
    :param y_columns:
    :param x_axis:
    :param epoch_col:
    :param profile:
    :param width:
    :param height:
    :param options:
    :param htmlCode:
    """
    epoch, labels, series = [], [], []
    for rec in record:
      epoch.append(rec[epoch_col])
      labels.append(rec[x_axis])
      series.append([rec.get(c)for c in y_columns])

    line_chart = graph.GraphC3.ChartStanford(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    line_chart.labels(labels)
    line_chart.epoch(epoch, epoch_col)
    for i, y in enumerate(y_columns):
      line_chart.add_dataset(y, series[i])
    return line_chart




import json

from epyk.core.js.html import JsHtml
from epyk.core.js.objects import JsNodeDom
from epyk.core.js import JsUtils
from epyk.core.css import Defaults

from epyk.core.js.primitives import JsObjects


class JsItemsDef(object):

  def _item(self, item_def):
    return '''%(item_def)s; htmlObj.appendChild(item)
      ''' % {'item_def': item_def}

  def text(self, report):
    """
    Description:
    ------------

    :param report:
    """
    item_def = '''
    var item = document.createElement("DIV");  
    item.setAttribute('name', 'value'); item.setAttribute('data-valid', true);
    if(options.click != null){ 
      item.style.cursor = 'pointer';
      item.onclick = function(event){ var value = this.innerHTML; options.click(event, value) }  };
    if(options.draggable != false){ 
      item.setAttribute('draggable', true);
      item.style.cursor = 'grab';
      item.ondragstart = function(event){ var value = this.innerHTML; options.draggable(event, value) }
    };
    if(typeof data === 'object'){ item.innerHTML = data.text} else { item.innerHTML = data }'''
    return self._item(item_def)

  def icon(self, report):
    """
    Description:
    ------------
    Set an list of icons

    Attributes:
    ----------
    :param report:
    """
    report.jsImports.add('font-awesome')
    report.cssImport.add('font-awesome')
    item_def = '''
    var item = document.createElement("DIV"); var icon = document.createElement("I"); 
    if(typeof data.icon !== 'undefined') {data.icon.split(" ").forEach(function(s){icon.classList.add(s)})}
    else {options.icon.split(" ").forEach(function(s){icon.classList.add(s)}) }
    icon.style.marginRight = '5px'; var span = document.createElement("span");  
    span.setAttribute('name', 'value'); span.setAttribute('data-valid', true); 
    if(typeof data === 'object'){ span.innerHTML = data.text} else { span.innerHTML = data };
    if(options.click != null){ item.style.cursor = 'pointer';
      item.onclick = function(event){ var value = span.innerHTML; options.click(event, value) }  };
    item.appendChild(icon); item.appendChild(span)'''
    return self._item(item_def)

  def check(self, report):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param report:
    """
    item_def = '''
    var item = document.createElement("DIV");  
    item.style.padding = 0; item.style.whiteSpace = "nowrap";  item.style.display = "block";
    
    var span = document.createElement("span");  
    var input = document.createElement("input");    
    input.setAttribute('type', 'checkbox');
    input.setAttribute('name', 'input_box');
    input.style.verticalAlign = "middle";
    input.onchange = function(event){ event.stopPropagation(); event.cancelBubble = true; span.setAttribute('data-valid', this.checked ); 
      var value = span.innerHTML; if(options.click != null){options.click(event, value)} };
    
    var span = document.createElement("span"); span.style.marginLeft = '5px'; 
    span.setAttribute('name', 'value'); span.setAttribute('data-valid', false); 
    
    if(typeof data === 'object'){ span.innerHTML = data.text} else { span.innerHTML = data };
    if(options.checked){ input.setAttribute('checked', options.checked); span.setAttribute('data-valid', options.checked) };
    if(data.checked){ input.setAttribute('checked', data.checked); span.setAttribute('data-valid', data.checked) };
    
    span.style.verticalAlign = "middle";
    item.appendChild(input); item.appendChild(span);
    '''
    return self._item(item_def)

  def radio(self, report):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param report:
    """
    item_def = '''
    var item = document.createElement("DIV");  
    item.style.padding = 0; item.style.whiteSpace = "nowrap";  item.style.display = "block";
    
    var span = document.createElement("span");  
    var input = document.createElement("input");    
    input.setAttribute('type', 'radio'); input.setAttribute('name', options.group);
    input.style.verticalAlign = "middle";
    input.onchange = function(event){ 
      this.parentNode.parentNode.parentNode.querySelectorAll('[name=value]').forEach(function(node){
        node.setAttribute('data-valid', false); }); var value = span.innerHTML; 
        span.setAttribute('data-valid', this.checked ); if(options.click != null){options.click(event, value)} };
    
    var span = document.createElement("span");  
    span.setAttribute('name', 'value'); span.setAttribute('data-valid', false);
    if(typeof data === 'object'){ span.innerHTML = data.text} else { span.innerHTML = data };
    span.style.verticalAlign = "middle";
    
    item.appendChild(input); item.appendChild(span);
    '''
    return self._item(item_def)

  def badge(self, report):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param report:
    """
    item_def = '''
    var item = document.createElement("DIV");  
    var span = document.createElement("span"); span.setAttribute('name', 'value'); span.innerHTML = data.text;  
    var badge = document.createElement("span"); badge.innerHTML = data.value;
    badge.style.backgroundColor = 'red'; badge.style.color = 'white'; badge.style.borderRadius = '50%%'; badge.style.padding = '0 3px';
    badge.style.marginLeft = '5px'; badge.style.fontSize = '%s'; 
    for(const attr in options.badge){badge.style[attr] = options.badge[attr]};
    item.appendChild(span); item.appendChild(badge)''' % Defaults.font(-2)
    return self._item(item_def)

  def link(self, report):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param report:
    """
    item_def = '''
    var item = document.createElement("a");  
    item.setAttribute('name', 'value'); item.setAttribute('data-valid', false);
    item.innerHTML = data ; item.href ='#' '''
    return self._item(item_def)

  def custom(self, item_def):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param item_def:
    """
    return self._item(item_def)


class JsItem(JsHtml.JsHtmlRich):

  @property
  def content(self):
    return JsHtml.ContentFormatters(self._report, '''
      (function(dom){var values = []; dom.childNodes.forEach( function(dom, k){  
          const valid = dom.querySelector('[name=value]').getAttribute("data-valid");
          if (valid === 'true'){values.push(dom.querySelector('[name=value]').innerHTML)}
      }); return values })(%s)''' % self.varName)

  @property
  def selected(self):
    """
    Description:
    ------------
    Return a list with all the selected values
    """
    return self.content

  @property
  def unselected(self):
    """
    Description:
    ------------
    Return a list with all the unselected values
    """
    return JsHtml.ContentFormatters(self._report, '''
        (function(dom){var values = []; dom.childNodes.forEach( function(dom, k){  
            const valid = dom.querySelector('[name=value]').getAttribute("data-valid");
            if (valid === 'false'){values.push(dom.querySelector('[name=value]').innerHTML)}
        }); return values })(%s)''' % self.varName)

  def getItemByValue(self, value):
    """
    Description:
    ------------
    Get an item from the list based on its value

    Attributes:
    ----------
    :param value: String. The value to find in the list
    """
    value = JsUtils.jsConvertData(value, None)
    return JsNodeDom.JsDoms.get('''
      (function(dom, value){var children = dom.childNodes; var values = null;
        for (var i = 0; i < children.length; i++) { if(children[i].querySelector('[name=value]').innerHTML == value){ values = children[i]; break} };
        return values })(%s, %s)''' % (self.varName, value))

  def selectAll(self, with_input_box=False):
    """
    Description:
    ------------
    Select all the items in the list

    Attributes:
    ----------
    :param with_input_box:
    """
    if self._src._jsStyles['items_type'] == "radio":
      raise Exception("It is not possible to select all radios from a same group, use check instead")

    if self._src._jsStyles['items_type'] == "check" or with_input_box:
      return JsObjects.JsVoid('''
        %s.childNodes.forEach( function(dom, k){  
          dom.querySelector('[name=input_box]').checked = true;
          dom.querySelector('[name=value]').setAttribute("data-valid", true);
        })''' % self.varName)

    return JsObjects.JsVoid('''
      %s.childNodes.forEach( function(dom, k){  
        dom.querySelector('[name=value]').setAttribute("data-valid", true);
      })''' % self.varName)

  def unSelectAll(self, with_input_box=False):
    """
    Description:
    ------------
    UnSelect all the items in the list

    Attributes:
    ----------
    :param with_input_box:
    """
    if self._src._jsStyles['items_type'] == "radio":
      raise Exception("It is not possible to select all radios from a same group, use check instead")

    if self._src._jsStyles['items_type'] == "check" or with_input_box:
      return JsObjects.JsVoid('''
        %s.childNodes.forEach( function(dom, k){  
          dom.querySelector('[name=input_box]').checked = false;
          dom.querySelector('[name=value]').setAttribute("data-valid", false);
        })''' % self.varName)

    return JsObjects.JsVoid('''
      %s.childNodes.forEach( function(dom, k){  
        dom.querySelector('[name=value]').setAttribute("data-valid", false);
      })''' % self.varName)

  def add(self, value):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param value: String | Dictionary.
    """
    if isinstance(value, dict):
      js_values = []
      for k, v in value.items():
        js_values.append("%s: %s" % (k, JsUtils.jsConvertData(v, None)))
      value = "{%s}" % ",".join(js_values)
    else:
      value = JsUtils.jsConvertData(value, None)
    return JsObjects.JsVoid('''
      var li = document.createElement("LI"); %(shape)s(li, %(value)s, %(options)s); li.style.margin = "5px 0"; %(comp)s.appendChild(li)
      ''' % {'comp': self.varName, 'options': json.dumps(self._src._jsStyles), 'value': value, 'shape': "%s%s" % (self._src._prefix, self._src._jsStyles['items_type'])})


class Tags(JsHtml.JsHtmlRich):

  @property
  def content(self):
    """
    Description:
    -----------
    Returns the list of data available on the filters panel
    """
    return JsHtml.ContentFormatters(self._report, '''
      (function(dom){var content = {}; 
        dom.childNodes.forEach(function(rec){
          var label = rec.getAttribute('data-category');
          if(!(label in content)){ content[label] = [] }; 
          content[label].push(rec.querySelector('span[name=chip_value]').textContent);
        }); 
        return content})(%s)
      ''' % self.querySelector("div[name=panel]"))

  def is_duplicated(self, text, category=None):
    """
    Description:
    ------------
    Check the duplicates in the filter panel for a given category

    Attributes:
    ----------
    :param text: String. The item text
    :param category: String. The item category
    """
    text = JsUtils.jsConvertData(text, None)
    return JsObjects.JsObjects.get(''' 
      (function(dom){var index = -1; var children = dom.childNodes; var count = 0;
        for(child in children){if((typeof children[child] === 'object') && children[child].querySelector('span[name=chip_value]').textContent == %s){
            if(children[child].getAttribute('data-category') == %s){ index = count; break; }
        }; count++; }; return index})(%s)''' % (text, category, self.querySelector("div[name=panel]")))

  def values(self, category=None):
    if category is None:
      return JsObjects.JsObjects.get("(function(dom){var content = []; dom.childNodes.forEach(function(rec){content.push(rec.querySelector('span[name=chip_value]').textContent)}); return content})(%s)" % self.querySelector("div[name=panel]"))

    category = JsUtils.jsConvertData(category, None)
    return JsObjects.JsObjects.get(''' 
          (function(dom){var children = dom.childNodes; var values = [];
            for(child in children){if(typeof children[child] === 'object'){
                if(children[child].getAttribute('data-category') == %s){ values.push(children[child].querySelector('span[name=chip_value]').textContent); }
            }}; return values})(%s)''' % (category, self.querySelector("div[name=panel]")))

  def hide(self):
    """
    Description:
    ------------
    Hide the filters panel
    """
    return self.querySelector("div[name=panel]").show()

  def show(self):
    """
    Description:
    ------------
    Show the filters panel
    """
    return self.querySelector("div[name=panel]").show()

  def toggle(self):
    """
    Description:
    ------------
    Toggle the display of the filters panel
    """
    return self.querySelector("div[name=panel]").toggle()

  def add(self, text, category=None, name=None, fixed=False, no_duplicte=True):
    """
    Description:
    ------------
    Add item on the filters panel

    Attributes:
    ----------
    :param text: String. The value to be added on the filter panel
    """
    text = JsUtils.jsConvertData(text, None)
    fixed = JsUtils.jsConvertData(fixed, None)
    if category is None:
      category = name or self._src._jsStyles['category']
    name = name or category
    category = JsUtils.jsConvertData(category, None)
    name = JsUtils.jsConvertData(name, None)
    # Convert the option to a javascript object
    # TODO move this in a centralised place
    options, js_options = self._src._jsStyles, []
    for k, v in options.items():
      if isinstance(v, dict):
        row = ["'%s': %s" % (s_k, JsUtils.jsConvertData(s_v, None)) for s_k, s_v in v.items()]
        js_options.append("'%s': {%s}" % (k, ", ".join(row)))
      else:
        if str(v).strip().startswith("function"):
          js_options.append("%s: %s" % (k, v))
        else:
          js_options.append("%s: %s" % (k, JsUtils.jsConvertData(v, None)))

    if no_duplicte:
      return JsObjects.JsObjects.get(''' 
      if ((%(duplicated)s == -1) && (%(text)s != '')){ chipAdd(%(panel)s, {name: %(name)s, category: %(category)s, value: %(text)s, disabled: false, fixed: %(fixed)s}, {%(options)s})  }
      ''' % {'name': name, 'category': category, 'duplicated': self.is_duplicated(text, category), 'panel': self.querySelector("div[name=panel]"), 'fixed': fixed, 'text': text, 'options': ",".join(js_options)})

    return JsObjects.JsObjects.get(''' 
      chipAdd(%(panel)s, {name: %(name)s, category: %(category)s, value: %(text)s, disabled: false, fixed: %(fixed)s}, {%(options)s})
      ''' % {'name': name, 'category': category, 'panel': self.querySelector("div[name=panel]"), 'fixed': fixed, 'text': text, 'options': ",".join(js_options)})

  @property
  def input(self):
    """
    Description:
    ------------
    Clear the content of the fitlers panel
    """
    return JsObjects.JsObjects.get("%s.value" % self.querySelector("input"))

  def clear(self):
    """
    Description:
    ------------
    Clear the content of the fitlers panel
    """
    return self.querySelector("div[name=panel]").empty()

  def remove(self, text, category=None):
    """
    Description:
    ------------
    Remove an item from the filters panel

    Attributes:
    ----------
    :param text: String. The test of the items to be removed
    :param category: String. The test of the items to be removed
    """
    if category is None:
      category = self._src._jsStyles['category']
    category = JsUtils.jsConvertData(category, None)
    return JsObjects.JsObjects.get('''var itemPos = %(duplicated)s; if (itemPos >= 0){ %(panel)s.childNodes[itemPos].remove()}
      ''' % {'duplicated': self.is_duplicated(text, category), 'panel': self.querySelector("div[name=panel]")})



"""
Wrapper to the Javascript Performance Module

The Performance interface provides access to performance-related information for the current page.
It's part of the High Resolution Time API, but is enhanced by the Performance Timeline API, the Navigation Timing API, the User Timing API, and the Resource Timing API

Related Pages:

			https://developer.mozilla.org/fr/docs/Web/API/Performance

"""


from epyk.core.js import JsUtils # All the predefined variable types
from epyk.core.js.fncs import JsFncs
from epyk.core.js.primitives import JsNumber
from epyk.core.js.primitives import JsObject
from epyk.core.js.primitives import JsArray


class JsPerformance(object):
  def __init__(self, src=None):
    self.__src = src
    self.__marks = set([])
    self.__count = 0

  def add_profiling(self, jsFnc):
    """
    Wrap the Javascript functions with function to asset on the execution time.

    Example
    self._report.js.performance.add_profiling(fncs['content'])

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/Performance/now

    :param jsFnc: The Javascript functions

    :return: The profile variable name
    """
    profile_var = "profile_%s" % self.__count
    if not isinstance(jsFnc, list):
      jsFnc = [jsFnc]
    jsFnc.insert(0, "var %s_start = %s" % (profile_var, self.now))
    jsFnc.append("var %s = %s - %s_start" % (profile_var, self.now, profile_var))
    self.__count += 1
    return profile_var

  def clearMarks(self, name=None):
    """
    The clearMarks() method removes the named mark from the browser's performance entry buffer.
    If the method is called with no arguments, all performance entries with an entry type of "mark" will be removed from the performance entry buffer.

    Example
    performance.clearMarks("a")

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/Performance/clearMarks

    :return: Void, the String for the Javascript side
    """
    if name is not None:
      if not name in self.__marks:
        raise Exception("Mark %s not defined in the performances" % name)

      return JsFncs.JsFunction("performance.clearMarks(%s)" % name)

    return JsFncs.JsFunction("performance.clearMarks()")

  def clearMeasures(self, name=None):
    """
    The clearMeasures() method removes the named measure from the browser's performance entry buffer.
    If the method is called with no arguments, all performance entries with an entry type of "measure" will be removed from the performance entry buffer.

    Example:
    performance.clearMeasures("a");

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/Performance/clearMeasures

    :param name: Optional, The name of the mark to be cleared

    :return: Void, the String for the Javascript side
    """
    if name is not None:
      self.__marks.remove(name)
      return JsFncs.JsFunction("performance.clearMeasures(%s)" % name)

    return JsFncs.JsFunction("performance.clearMeasures()")

  def clearResourceTimings(self):
    """
    The clearResourceTimings() method removes all performance entries with an entryType of "resource" from the browser's performance data buffer and sets the size of the performance data buffer to zero.
    To set the size of the browser's performance data buffer, use the Performance.setResourceTimingBufferSize() method.

    Example
    performance.clearResourceTimings()

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/Performance/clearResourceTimings

    :return: This method has no return value, but only the String for the Javascript side
    """
    return JsFncs.JsFunction("performance.clearResourceTimings()")

  def getEntries(self):
    """
    The getEntries() method returns a list of all PerformanceEntry objects for the page.
    The list's members (entries) can be created by making performance marks or measures (for example by calling the mark() method) at explicit points in time.
    If you are only interested in performance entries of certain types or that have certain names, see getEntriesByType() and getEntriesByName().

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/Performance/getEntries

    :return: An array of PerformanceEntry objects
    """
    return JsArray.JsArray("window.performance.getEntries()", isPyData=False)

  def getEntriesByName(self, name, type=None):
    """
    The getEntriesByName() method returns a list of PerformanceEntry objects for the given name and type.
    The list's members (entries) can be created by making performance marks or measures (for example by calling the mark() method) at explicit points in time.

    Example
    performance.getEntriesByName("Begin", "mark")

    Related Pages:

			https//developer.mozilla.org/en-US/docs/Web/API/Performance/getEntriesByName

    :param name: The name of the entry to retrieve.
    :param type: Optional, The type of entry to retrieve such as "mark"

    :return: A list of PerformanceEntry objects that have the specified name and type
    """
    if not name in self.__marks:
      raise Exception("Mark %s not defined in the performances" % name)

    name = JsUtils.jsConvertData(name, None)
    if type is not None:
      return JsArray.JsArray("window.performance.getEntriesByName(%s, %s)" % (name, type), isPyData=False)

    return JsArray.JsArray("window.performance.getEntriesByName(%s)" % name, isPyData=False)

  def getEntriesByType(self, type):
    """
    The getEntriesByType() method returns a list of PerformanceEntry objects for a given type.
    The list's members (entries) can be created by making performance marks or measures (for example by calling the mark() method) at explicit points in time.

    Example
    performance.getEntriesByType("mark")

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/Performance/getEntriesByType

    :param type: The type of entry to retrieve such as "mark"

    :return: A list of PerformanceEntry objects that have the specified type
    """
    return JsArray.JsArray("window.performance.getEntriesByType('%s')" % type, isPyData=False)

  def mark(self, name):
    """
    The mark() method creates a timestamp in the browser's performance entry buffer with the given name.
    The application defined timestamp can be retrieved by one of the Performance interface's getEntries*() methods (getEntries(), getEntriesByName() or getEntriesByType()).

    Example:
    performance.mark("a")

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/Performance/mark

    :param name: A DOMString representing the name of the mark.

    :return: VOid, The String for the Javascript side
    """
    self.__marks.add(name)
    name = JsUtils.jsConvertData(name, None)
    return JsFncs.JsFunction("performance.mark(%s)" % name)

  def measure(self, name, startMark=None, endMark=None):
    """
    The measure() method creates a named timestamp in the browser's performance entry buffer between marks, the navigation start time, or the current time.
    When measuring between two marks, there is a start mark and end mark, respectively.
    The named timestamp is referred to as a measure.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/Performance/measure

    :param name: A DOMString representing the name of the measure
    :param startMark: Optional, A DOMString representing the name of the measure's starting mark
    :param endMark: Optional, A DOMString representing the name of the measure's ending mark

    :return: Void, The String for the Javascript side
    """
    name = JsUtils.jsConvertData(name, None)
    if startMark is not None:
      if not startMark in self.__marks:
        raise Exception("Mark %s not defined in the performances" % startMark)

      if endMark is not None:
        if not startMark in self.__marks:
          raise Exception("Mark %s not defined in the performances" % startMark)

        return JsFncs.JsFunction("performance.measure(%s, %s, %s)" % (name, startMark, endMark))
      else:
        return JsFncs.JsFunction("performance.measure(%s, '%s')" % (name, startMark))

    return JsFncs.JsFunction("performance.measure(%s)" % name)

  @property
  def now(self):
    """
    The performance.now() method returns a DOMHighResTimeStamp, measured in milliseconds.

    Example
    var t0 = performance.now();

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/Performance/now

    :return: A Javascript Number
    """
    return JsNumber.JsNumber("performance.now()", isPyData=False)

  def setResourceTimingBufferSize(self, maxSize):
    """
    The setResourceTimingBufferSize() method sets the browser's resource timing buffer size to the specified number of "resource" performance entry type objects.

    Example
    performance.setResourceTimingBufferSize(maxSize)

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/Performance/setResourceTimingBufferSize

    :return: Void, the String for the Javascript side
    """
    return JsFncs.JsFunction("performance.setResourceTimingBufferSize(%s)" % maxSize)

  def toJSON(self):
    """
    The toJSON() method of the Performance interface is a standard serializer: it returns a JSON representation of the performance object's properties.

    Example
    performance.toJSON()

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/Performance/toJSON

    :return: A JSON object that is the serialization of the Performance object.
    """
    return JsObject.JsObject("performance.toJSON()", isPyData=False)





from epyk.core.html import Html

from epyk.core.js.packages import JsDc


class Chart(Html.Html):
  name = 'DC Chart'
  requirements = ('dc', 'crossfilter')

  def __init__(self,  report, width, height, title, options, htmlCode, profile):
    super(Chart, self).__init__(report, [], htmlCode=htmlCode, css_attrs={"width": width, "height": height}, profile=profile)
    self.style.css.margin = "10px 0"

  @property
  def chartId(self):
    """
    Return the Javascript variable of the chart
    """
    return "chart_%s" % self.htmlCode

  def crossFilter(self, dimension, group):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param dimension:
    :param group:
    """
    self.dom.dimension(dimension.varId).group(group.varId)
    return self

  def build(self, data=None, options=None, profile=False):
    return self.dom.render().toStr()

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    return '<div %s></div>' % self.get_attrs(pyClassNames=self.style.get_classes())


class ChartLine(Chart):

  @property
  def dom(self):
    """
    Description:
    -----------
    A line chart is used to display information as a series of data points connected by straight lines.
    A data point represents two values, one plotted along the horizontal axis and another along the vertical axis.
    For example, the popularity of food items can be drawn as a line chart in such a way that the food item is represented along the x-axis and its popularity is represented along the y-axis.

    Related Pages:

			https://www.tutorialspoint.com/dcjs/dcjs_line_chart.htm

    :rtype: JsChartDC.JsLine
    """
    if self._dom is None:
      self._dom = JsDc.Line(self._report, varName=self.chartId, parent=self)
    return self._dom


class ChartBar(Chart):

  @property
  def dom(self):
    """
    Description:
    -----------
    Bar chart is one of the most commonly used types of graph and are used to display and compare the number, frequency or other measure (e.g. mean) for different discrete categories or groups.
    The graph is constructed such that the heights or lengths of the different bars are proportional to the size of the category they represent.

    Related Pages:

			https://www.tutorialspoint.com/dcjs/dcjs_bar_chart.htm

    :rtype: JsDc.Bar
    """
    if self._dom is None:
      self._dom = JsDc.Bar(self._report, varName=self.chartId, parent=self)
    return self._dom


class ChartRow(Chart):

  @property
  def dom(self):
    """
    :rtype: JsDc.Row
    """
    if self._dom is None:
      self._dom = JsDc.Row(self._report, varName=self.chartId, parent=self)
    return self._dom


class ChartScatter(Chart):

  @property
  def dom(self):
    """
    Description:
    -----------
    A scatter plot is a type of mathematical diagram.
    It is represented using the Cartesian coordinates to display values for typically two variables for a set of data.
    The data is displayed as a collection of points and the points maybe colored.

    Related Pages:

			https://www.tutorialspoint.com/dcjs/dcjs_scatter_plot.htm

    :rtype: JsDc.Scatter
    """
    if self._dom is None:
      self._dom = JsDc.Scatter(self._report, varName=self.chartId, parent=self)
    return self._dom


class ChartBubble(Chart):

  @property
  def dom(self):
    """
    Description:
    -----------
    A bubble chart is used to display three dimensions of the data.
    It is a variation of scatter chart, in which the data points are replaced with bubbles. The bubble sizes are represented with respect to the data dimension.
    It uses horizontal and vertical axes as value axes.

    Related Pages:

			https://www.tutorialspoint.com/dcjs/dcjs_bubble_chart.htm

    :rtype: JsDc.Bubble
    """
    if self._dom is None:
      self._dom = JsDc.Bubble(self._report, varName=self.chartId, parent=self)
    return self._dom


class ChartPie(Chart):

  @property
  def dom(self):
    """
    Description:
    -----------
    A pie chart is a circular statistical graph. It is divided into slices to show a numerical proportion

    Related Pages:

			https://www.tutorialspoint.com/dcjs/dcjs_pie_chart.htm

    :rtype: JsDc.Pie
    """
    if self._dom is None:
      self._dom = JsDc.Pie(self._report, varName=self.chartId, parent=self)
    return self._dom


class ChartSunburst(Chart):

  @property
  def dom(self):
    """
    :rtype: JsDc.Sunburst
    """
    if self._dom is None:
      self._dom = JsDc.Sunburst(self._report, varName=self.chartId, parent=self)
    return self._dom


class ChartSeries(Chart):

  @property
  def dom(self):
    """
    Description:
    -----------
    A series is a set of data. You can plot a chart based on the data.

    Related Pages:

			https://www.tutorialspoint.com/dcjs/dcjs_series_chart.htm

    :rtype: JsDc.Series
    """
    if self._dom is None:
      self._dom = JsDc.Series(self._report, varName=self.chartId, parent=self)
    return self._dom
